@@ -13,7 +13,8 @@ Wired-in knowledge about primitive types
 module GHC.Builtin.Types.Prim(
         mkPrimTyConName, -- For implicit parameters in GHC.Builtin.Types only
 
-        mkTemplateKindVars, mkTemplateTyVars, mkTemplateTyVarsFrom,
+        mkTemplateKindVar, mkTemplateKindVars,
+        mkTemplateTyVars, mkTemplateTyVarsFrom,
         mkTemplateKiTyVars, mkTemplateKiTyVar,
 
         mkTemplateTyConBinders, mkTemplateKindTyConBinders,
@@ -35,7 +36,7 @@ module GHC.Builtin.Types.Prim(
         tYPETyCon, tYPETyConName,
 
         -- Kinds
-        tYPE, primRepToRuntimeRep,
+        tYPE, primRepToRuntimeRep, primRepsToRuntimeRep,
 
         functionWithMultiplicity,
         funTyCon, funTyConName,
@@ -587,7 +588,7 @@ pcPrimTyCon name roles rep
 -- Defined here to avoid (more) module loops
 primRepToRuntimeRep :: PrimRep -> Type
 primRepToRuntimeRep rep = case rep of
-  VoidRep       -> TyConApp tupleRepDataConTyCon [mkPromotedListTy runtimeRepTy []]
+  VoidRep       -> mkTupleRep []
   LiftedRep     -> liftedRepTy
   UnliftedRep   -> unliftedRepTy
   IntRep        -> intRepDataConTy
@@ -626,6 +627,17 @@ primRepToRuntimeRep rep = case rep of
         FloatElemRep  -> floatElemRepDataConTy
         DoubleElemRep -> doubleElemRepDataConTy
 
+-- | Given a list of types representing 'RuntimeRep's @reps@, construct
+-- @'TupleRep' reps@.
+mkTupleRep :: [Type] -> Type
+mkTupleRep reps = TyConApp tupleRepDataConTyCon [mkPromotedListTy runtimeRepTy reps]
+
+-- | Convert a list of 'PrimRep's to a 'Type' of kind RuntimeRep
+-- Defined here to avoid (more) module loops
+primRepsToRuntimeRep :: [PrimRep] -> Type
+primRepsToRuntimeRep [rep] = primRepToRuntimeRep rep
+primRepsToRuntimeRep reps  = mkTupleRep $ map primRepToRuntimeRep reps
+
 pcPrimTyCon0 :: Name -> PrimRep -> TyCon
 pcPrimTyCon0 name rep
   = pcPrimTyCon name [] rep
