@@ -14,33 +14,51 @@ where
 
 import GHC.Prelude
 
+import GHC.Driver.Backend
+import GHC.Driver.Env
+import GHC.Driver.Session
+import GHC.Driver.Ppr
+import GHC.Driver.Plugins ( PluginRecompile(..), PluginWithArgs(..), pluginRecompile', plugins )
+
 import GHC.Iface.Syntax
 import GHC.Iface.Recomp.Binary
 import GHC.Iface.Load
 import GHC.Iface.Recomp.Flags
 
-import GHC.Types.Annotations
 import GHC.Core
 import GHC.Tc.Utils.Monad
 import GHC.Hs
-import GHC.Driver.Types
-import GHC.Driver.Finder
-import GHC.Driver.Session
-import GHC.Types.Name
-import GHC.Types.Name.Set
-import GHC.Unit.Module
-import GHC.Utils.Error
+
 import GHC.Data.Graph.Directed
-import GHC.Types.SrcLoc
-import GHC.Utils.Outputable as Outputable
-import GHC.Types.Unique
-import GHC.Utils.Misc as Utils hiding ( eqListBy )
 import GHC.Data.Maybe
+import GHC.Data.FastString
+
+import GHC.Utils.Error
+import GHC.Utils.Panic
+import GHC.Utils.Outputable as Outputable
+import GHC.Utils.Misc as Utils hiding ( eqListBy )
 import GHC.Utils.Binary
 import GHC.Utils.Fingerprint
 import GHC.Utils.Exception
+
+import GHC.Types.Annotations
+import GHC.Types.Name
+import GHC.Types.Name.Set
+import GHC.Types.SrcLoc
+import GHC.Types.Unique
 import GHC.Types.Unique.Set
+import GHC.Types.Fixity.Env
+import GHC.Types.SourceFile
+
+import GHC.Unit.External
+import GHC.Unit.Finder
 import GHC.Unit.State
+import GHC.Unit.Home
+import GHC.Unit.Module
+import GHC.Unit.Module.ModIface
+import GHC.Unit.Module.ModSummary
+import GHC.Unit.Module.Warnings
+import GHC.Unit.Module.Deps
 
 import Control.Monad
 import Data.Function
@@ -48,7 +66,6 @@ import Data.List (find, sortBy, sort)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 import Data.Word (Word64)
-import GHC.Driver.Plugins ( PluginRecompile(..), PluginWithArgs(..), pluginRecompile', plugins )
 
 --Qualified import so we can define a Semigroup instance
 -- but it doesn't clash with Outputable.<>
@@ -122,7 +139,8 @@ checkOldIface
 
 checkOldIface hsc_env mod_summary source_modified maybe_iface
   = do  let dflags = hsc_dflags hsc_env
-        showPass dflags $
+        let logger = hsc_logger hsc_env
+        showPass logger dflags $
             "Checking old interface for " ++
               (showPpr dflags $ ms_mod mod_summary) ++
               " (use -ddump-hi-diffs for more details)"
@@ -170,7 +188,7 @@ check_old_iface hsc_env mod_summary src_modified maybe_iface
             -- If the source has changed and we're in interactive mode,
             -- avoid reading an interface; just return the one we might
             -- have been supplied with.
-            True | not (isObjectTarget $ hscTarget dflags) ->
+            True | not (backendProducesObject $ backend dflags) ->
                 return (MustCompile, maybe_iface)
 
             -- Try and read the old interface for the current module
@@ -213,7 +231,7 @@ checkVersions hsc_env mod_summary iface
        -- readIface will have verified that the UnitId matches,
        -- but we ALSO must make sure the instantiation matches up.  See
        -- test case bkpcabal04!
-       ; if moduleUnit (mi_module iface) /= homeUnit (hsc_dflags hsc_env)
+       ; if not (isHomeModule home_unit (mi_module iface))
             then return (RecompBecause "-this-unit-id changed", Nothing) else do {
        ; recomp <- checkFlagHash hsc_env iface
        ; if recompileRequired recomp then return (recomp, Nothing) else do {
@@ -247,27 +265,28 @@ checkVersions hsc_env mod_summary iface
        -- all the dependent modules should be in the HPT already, so it's
        -- quite redundant
        ; updateEps_ $ \eps  -> eps { eps_is_boot = mod_deps }
-       ; recomp <- checkList [checkModUsage this_pkg u | u <- mi_usages iface]
+       ; recomp <- checkList [checkModUsage (homeUnitAsUnit home_unit) u
+                             | u <- mi_usages iface]
        ; return (recomp, Just iface)
     }}}}}}}}}}
   where
-    this_pkg = homeUnit (hsc_dflags hsc_env)
+    home_unit = hsc_home_unit hsc_env
     -- This is a bit of a hack really
     mod_deps :: ModuleNameEnv ModuleNameWithIsBoot
     mod_deps = mkModDeps (dep_mods (mi_deps iface))
 
 -- | Check if any plugins are requesting recompilation
 checkPlugins :: HscEnv -> ModIface -> IfG RecompileRequired
-checkPlugins hsc iface = liftIO $ do
-  new_fingerprint <- fingerprintPlugins hsc
+checkPlugins hsc_env iface = liftIO $ do
+  new_fingerprint <- fingerprintPlugins hsc_env
   let old_fingerprint = mi_plugin_hash (mi_final_exts iface)
-  pr <- mconcat <$> mapM pluginRecompile' (plugins (hsc_dflags hsc))
+  pr <- mconcat <$> mapM pluginRecompile' (plugins hsc_env)
   return $
     pluginRecompileToRecompileRequired old_fingerprint new_fingerprint pr
 
 fingerprintPlugins :: HscEnv -> IO Fingerprint
-fingerprintPlugins hsc_env = do
-  fingerprintPlugins' $ plugins (hsc_dflags hsc_env)
+fingerprintPlugins hsc_env =
+  fingerprintPlugins' $ plugins hsc_env
 
 fingerprintPlugins' :: [PluginWithArgs] -> IO Fingerprint
 fingerprintPlugins' plugins = do
@@ -330,10 +349,11 @@ pluginRecompileToRecompileRequired old_fp new_fp pr
 -- implementing module has changed.
 checkHsig :: ModSummary -> ModIface -> IfG RecompileRequired
 checkHsig mod_summary iface = do
-    dflags <- getDynFlags
-    let outer_mod = ms_mod mod_summary
-        inner_mod = canonicalizeHomeModule dflags (moduleName outer_mod)
-    MASSERT( moduleUnit outer_mod == homeUnit dflags )
+    hsc_env <- getTopEnv
+    let home_unit = hsc_home_unit hsc_env
+        outer_mod = ms_mod mod_summary
+        inner_mod = homeModuleNameInstantiation home_unit (moduleName outer_mod)
+    MASSERT( isHomeModule home_unit outer_mod )
     case inner_mod == mi_semantic_module iface of
         True -> up_to_date (text "implementing module unchanged")
         False -> return (RecompBecause "implementing module changed")
@@ -358,7 +378,7 @@ checkHie mod_summary = do
 checkFlagHash :: HscEnv -> ModIface -> IfG RecompileRequired
 checkFlagHash hsc_env iface = do
     let old_hash = mi_flag_hash (mi_final_exts iface)
-    new_hash <- liftIO $ fingerprintDynFlags (hsc_dflags hsc_env)
+    new_hash <- liftIO $ fingerprintDynFlags hsc_env
                                              (mi_module iface)
                                              putNameLiterally
     case old_hash == new_hash of
@@ -401,12 +421,12 @@ checkHpcHash hsc_env iface = do
 -- If the -unit-id flags change, this can change too.
 checkMergedSignatures :: ModSummary -> ModIface -> IfG RecompileRequired
 checkMergedSignatures mod_summary iface = do
-    dflags <- getDynFlags
+    unit_state <- hsc_units <$> getTopEnv
     let old_merged = sort [ mod | UsageMergedRequirement{ usg_mod = mod } <- mi_usages iface ]
         new_merged = case Map.lookup (ms_mod_name mod_summary)
-                                     (requirementContext (unitState dflags)) of
+                                     (requirementContext unit_state) of
                         Nothing -> []
-                        Just r -> sort $ map (instModuleToModule (unitState dflags)) r
+                        Just r -> sort $ map (instModuleToModule unit_state) r
     if old_merged == new_merged
         then up_to_date (text "signatures to merge in unchanged" $$ ppr new_merged)
         else return (RecompBecause "signatures to merge in changed")
@@ -430,7 +450,7 @@ checkMergedSignatures mod_summary iface = do
 -- Returns (RecompBecause <textual reason>) if recompilation is required.
 checkDependencies :: HscEnv -> ModSummary -> ModIface -> IfG RecompileRequired
 checkDependencies hsc_env summary iface
- = do
+ =
    checkList $
      [ checkList (map dep_missing (ms_imps summary ++ ms_srcimps summary))
      , do
@@ -447,15 +467,14 @@ checkDependencies hsc_env summary iface
    prev_dep_mods = dep_mods (mi_deps iface)
    prev_dep_plgn = dep_plgins (mi_deps iface)
    prev_dep_pkgs = dep_pkgs (mi_deps iface)
-
-   this_pkg = homeUnit (hsc_dflags hsc_env)
+   home_unit     = hsc_home_unit hsc_env
 
    dep_missing (mb_pkg, L _ mod) = do
      find_res <- liftIO $ findImportedModule hsc_env mod (mb_pkg)
      let reason = moduleNameString mod ++ " changed"
      case find_res of
         Found _ mod
-          | pkg == this_pkg
+          | isHomeUnit home_unit pkg
            -> if moduleName mod `notElem` map gwib_mod prev_dep_mods ++ prev_dep_plgn
                  then do traceHiDiffs $
                            text "imported module " <> quotes (ppr mod) <>
@@ -481,7 +500,7 @@ checkDependencies hsc_env summary iface
    isOldHomeDeps = flip Set.member old_deps
    checkForNewHomeDependency (L _ mname) = do
      let
-       mod = mkModule this_pkg mname
+       mod = mkHomeModule home_unit mname
        str_mname = moduleNameString mname
        reason = str_mname ++ " changed"
      -- We only want to look at home modules to check if any new home dependency
@@ -601,8 +620,7 @@ checkModUsage this_pkg UsageHomeModule{
        recompile <- checkModuleFingerprint reason old_mod_hash new_mod_hash
        if not (recompileRequired recompile)
          then return UpToDate
-         else do
-
+         else
            -- CHECK EXPORT LIST
            checkMaybeHash reason maybe_old_export_hash new_export_hash
                (text "  Export list changed") $ do
@@ -937,7 +955,7 @@ addFingerprints hsc_env iface0
        extend_hash_env :: OccEnv (OccName,Fingerprint)
                        -> (Fingerprint,IfaceDecl)
                        -> IO (OccEnv (OccName,Fingerprint))
-       extend_hash_env env0 (hash,d) = do
+       extend_hash_env env0 (hash,d) =
           return (foldr (\(b,fp) env -> extendOccEnv env b (b,fp)) env0
                  (ifaceDeclFingerprints hash d))
 
@@ -948,7 +966,8 @@ addFingerprints hsc_env iface0
    -- when calculating fingerprints, we always need to use canonical
    -- ordering for lists of things.  In particular, the mi_deps has various
    -- lists of modules and suchlike, so put these all in canonical order:
-   let sorted_deps = sortDependencies (mi_deps iface0)
+   let sorted_deps :: Dependencies
+       sorted_deps = sortDependencies (mi_deps iface0)
 
    -- The export hash of a module depends on the orphan hashes of the
    -- orphan modules below us in the dependency tree.  This is the way
@@ -1043,7 +1062,7 @@ addFingerprints hsc_env iface0
    --   - (some of) dflags
    -- it returns two hashes, one that shouldn't change
    -- the abi hash and one that should
-   flag_hash <- fingerprintDynFlags dflags this_mod putNameLiterally
+   flag_hash <- fingerprintDynFlags hsc_env this_mod putNameLiterally
 
    opt_hash <- fingerprintOptFlags dflags putNameLiterally
 
@@ -1157,11 +1176,11 @@ getOrphanHashes hsc_env mods = do
 
 sortDependencies :: Dependencies -> Dependencies
 sortDependencies d
- = Deps { dep_mods   = sortBy (compare `on` (moduleNameFS . gwib_mod)) (dep_mods d),
+ = Deps { dep_mods   = sortBy (lexicalCompareFS `on` (moduleNameFS . gwib_mod)) (dep_mods d),
           dep_pkgs   = sortBy (compare `on` fst) (dep_pkgs d),
           dep_orphs  = sortBy stableModuleCmp (dep_orphs d),
           dep_finsts = sortBy stableModuleCmp (dep_finsts d),
-          dep_plgins = sortBy (compare `on` moduleNameFS) (dep_plgins d) }
+          dep_plgins = sortBy (lexicalCompareFS `on` moduleNameFS) (dep_plgins d) }
 
 {-
 ************************************************************************
@@ -1430,11 +1449,11 @@ mkHashFun
         -> (Name -> IO Fingerprint)
 mkHashFun hsc_env eps name
   | isHoleModule orig_mod
-  = lookup (mkHomeModule dflags (moduleName orig_mod))
+  = lookup (mkHomeModule home_unit (moduleName orig_mod))
   | otherwise
   = lookup orig_mod
   where
-      dflags = hsc_dflags hsc_env
+      home_unit = hsc_home_unit hsc_env
       hpt = hsc_HPT hsc_env
       pit = eps_PIT eps
       occ = nameOccName name
@@ -1443,14 +1462,18 @@ mkHashFun hsc_env eps name
         MASSERT2( isExternalName name, ppr name )
         iface <- case lookupIfaceByModule hpt pit mod of
                   Just iface -> return iface
-                  Nothing -> do
+                  Nothing ->
                       -- This can occur when we're writing out ifaces for
                       -- requirements; we didn't do any /real/ typechecking
                       -- so there's no guarantee everything is loaded.
                       -- Kind of a heinous hack.
-                      iface <- initIfaceLoad hsc_env . withException
-                            $ loadInterface (text "lookupVers2") mod ImportBySystem
-                      return iface
+                      initIfaceLoad hsc_env . withException
+                          $ withoutDynamicNow
+                            -- For some unknown reason, we need to reset the
+                            -- dynamicNow bit, otherwise only dynamic
+                            -- interfaces are looked up and some tests fail
+                            -- (e.g. T16219).
+                          $ loadInterface (text "lookupVers2") mod ImportBySystem
         return $ snd (mi_hash_fn (mi_final_exts iface) occ `orElse`
                   pprPanic "lookupVers1" (ppr mod <+> ppr occ))
 
