@@ -1421,6 +1421,7 @@ tcIfaceUnivCoProv :: IfaceUnivCoProv -> IfL UnivCoProvenance
 tcIfaceUnivCoProv (IfacePhantomProv kco)    = PhantomProv <$> tcIfaceCo kco
 tcIfaceUnivCoProv (IfaceProofIrrelProv kco) = ProofIrrelProv <$> tcIfaceCo kco
 tcIfaceUnivCoProv (IfacePluginProv str)     = return $ PluginProv str
+tcIfaceUnivCoProv (IfaceCorePrepProv b)     = return $ CorePrepProv b
 
 {-
 ************************************************************************
@@ -1446,6 +1447,10 @@ tcIfaceExpr (IfaceLcl name)
 tcIfaceExpr (IfaceExt gbl)
   = Var <$> tcIfaceExtId gbl
 
+tcIfaceExpr (IfaceLitRubbish rep)
+  = do rep' <- tcIfaceType rep
+       return (Lit (LitRubbish rep'))
+
 tcIfaceExpr (IfaceLit lit)
   = do lit' <- tcIfaceLit lit
        return (Lit lit')
@@ -1536,9 +1541,9 @@ tcIfaceExpr (IfaceLet (IfaceRec pairs) body)
 tcIfaceExpr (IfaceTick tickish expr) = do
     expr' <- tcIfaceExpr expr
     -- If debug flag is not set: Ignore source notes
-    dbgLvl <- fmap debugLevel getDynFlags
+    need_notes <- needSourceNotes <$> getDynFlags
     case tickish of
-      IfaceSource{} | dbgLvl == 0
+      IfaceSource{} | not (need_notes)
                     -> return expr'
       _otherwise    -> do
         tickish' <- tcIfaceTickish tickish
@@ -1742,7 +1747,7 @@ tcPragExpr is_compulsory toplvl name expr
     core_expr' <- tcIfaceExpr expr
 
     -- Check for type consistency in the unfolding
-    -- See Note [Linting Unfoldings from Interfaces]
+    -- See Note [Linting Unfoldings from Interfaces] in GHC.Core.Lint
     when (isTopLevel toplvl) $
       whenGOptM Opt_DoCoreLinting $ do
         in_scope <- get_in_scope
