@@ -218,6 +218,11 @@ platformCConvNeedsExtension platform = case platformArch platform of
   ArchPPC_64 _ -> True
   ArchS390X    -> True
   ArchRISCV64  -> True
+  ArchAArch64
+      -- Apple's AArch64 ABI requires that the caller sign-extend
+      -- small integer arguments. See
+      -- https://developer.apple.com/documentation/xcode/writing-arm64-code-for-apple-platforms
+    | OSDarwin <- platformOS platform -> True
   _            -> False
 
 
