@@ -2,7 +2,7 @@
 
 module GHC.Platform.PPC where
 
-import GhcPrelude
+import GHC.Prelude
 
 #define MACHREGS_NO_REGS 0
 #define MACHREGS_powerpc 1
