@@ -157,17 +157,18 @@ Quick Look treats application chains specially.  What is an
 "application chain"?  See Fig 2, of the QL paper: "A quick look at
 impredicativity" (ICFP'20). Here's the syntax:
 
-app :: head
-     | app expr            -- HsApp: ordinary application
-     | app @type           -- HsTypeApp: VTA
-     | expr `head` expr    -- OpApp: infix applications
-     | ( app )             -- HsPar: parens
-     | {-# PRAGMA #-} app  -- HsPragE: pragmas
+app ::= head
+     |  app expr            -- HsApp: ordinary application
+     |  app @type           -- HsTypeApp: VTA
+     |  expr `head` expr    -- OpApp: infix applications
+     |  ( app )             -- HsPar: parens
+     |  {-# PRAGMA #-} app  -- HsPragE: pragmas
 
 head ::= f             -- HsVar:    variables
       |  fld           -- HsRecFld: record field selectors
       |  (expr :: ty)  -- ExprWithTySig: expr with user type sig
       |  lit           -- HsOverLit: overloaded literals
+      |  $([| head |])    -- HsSpliceE+HsSpliced+HsSplicedExpr: untyped TH expression splices
       |  other_expr    -- Other expressions
 
 When tcExpr sees something that starts an application chain (namely,
@@ -185,7 +186,7 @@ There is no special treatment for HsUnboundVar, HsOverLit etc, because
 we can't get a polytype from them.
 
 Left and right sections (e.g. (x +) and (+ x)) are not yet supported.
-Probably left sections (x +) would be esay to add, since x is the
+Probably left sections (x +) would be easy to add, since x is the
 first arg of (+); but right sections are not so easy.  For symmetry
 reasons I've left both unchanged, in GHC.Tc.Gen.Expr.
 
@@ -197,6 +198,16 @@ Clearly this should work!  But it will /only/ work because if we
 instantiate that (forall b. b) impredicatively!  And that only happens
 in tcApp.
 
+We also wish to typecheck application chains with untyped Template Haskell
+splices in the head, such as this example from #21038:
+    data Foo = MkFoo (forall a. a -> a)
+    f = $([| MkFoo |]) $ \x -> x
+This should typecheck just as if the TH splice was never in the way—that is,
+just as if the user had written `MkFoo $ \x -> x`. We could conceivably have
+a case for typed TH expression splices too, but it wouldn't be useful in
+practice, since the types of typed TH expressions aren't allowed to have
+polymorphic types, such as the type of MkFoo.
+
 Note [tcApp: typechecking applications]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 tcApp implements the APP-Downarrow/Uparrow rule of
@@ -278,6 +289,7 @@ particular Ids:
   the renamer (Note [Handling overloaded and rebindable constructs] in
   GHC.Rename.Expr), and we want them to be instantiated impredicatively
   so that (f `op`), say, will work OK even if `f` is higher rank.
+  See Note [Left and right sections] in GHC.Rename.Expr.
 
 Note [Unify with expected type before typechecking arguments]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -338,8 +350,26 @@ tcApp rn_expr exp_res_ty
                  = addFunResCtxt rn_fun rn_args app_res_rho exp_res_ty $
                    thing_inside
 
-       ; res_co <- perhaps_add_res_ty_ctxt $
-                   unifyExpectedType rn_expr app_res_rho exp_res_ty
+       -- Match up app_res_rho: the result type of rn_expr
+       --     with exp_res_ty:  the expected result type
+       ; do_ds <- xoptM LangExt.DeepSubsumption
+       ; res_wrap <- perhaps_add_res_ty_ctxt $
+            if not do_ds
+            then -- No deep subsumption
+                 -- app_res_rho and exp_res_ty are both rho-types,
+                 -- so with simple subsumption we can just unify them
+                 -- No need to zonk; the unifier does that
+                 do { co <- unifyExpectedType rn_expr app_res_rho exp_res_ty
+                    ; return (mkWpCastN co) }
+
+            else -- Deep subsumption
+                 -- Even though both app_res_rho and exp_res_ty are rho-types,
+                 -- they may have nested polymorphism, so if deep subsumption
+                 -- is on we must call tcSubType.
+                 -- Zonk app_res_rho first, becuase QL may have instantiated some
+                 -- delta variables to polytypes, and tcSubType doesn't expect that
+                 do { app_res_rho <- zonkQuickLook do_ql app_res_rho
+                    ; tcSubTypeDS rn_expr app_res_rho exp_res_ty }
 
        ; whenDOptM Opt_D_dump_tc_trace $
          do { inst_args <- mapM zonkArg inst_args  -- Only when tracing
@@ -361,11 +391,10 @@ tcApp rn_expr exp_res_ty
                     else return (rebuildHsApps tc_fun fun_ctxt tc_args)
 
        -- Wrap the result
-       ; return (mkHsWrapCo res_co tc_expr) }
+       ; return (mkHsWrap res_wrap tc_expr) }
 
 --------------------
 wantQuickLook :: HsExpr GhcRn -> TcM Bool
--- GHC switches on impredicativity all the time for ($)
 wantQuickLook (HsVar _ (L _ f))
   | getUnique f `elem` quickLookKeys = return True
 wantQuickLook _                      = xoptM LangExt.ImpredicativeTypes
