@@ -698,8 +698,7 @@ writeTcRef = writeMutVar
 
 updTcRef :: TcRef a -> (a -> a) -> TcRnIf gbl lcl ()
 -- Returns ()
-updTcRef ref fn = liftIO $ do { old <- readIORef ref
-                              ; writeIORef ref (fn old) }
+updTcRef ref fn = liftIO $ modifyIORef' ref fn
 
 {-
 ************************************************************************
