@@ -20,15 +20,15 @@ module GHC.Types.Literal
 
         -- ** Creating Literals
         , mkLitInt, mkLitIntWrap, mkLitIntWrapC, mkLitIntUnchecked
-        , mkLitWord, mkLitWordWrap, mkLitWordWrapC
-        , mkLitInt8, mkLitInt8Wrap
-        , mkLitWord8, mkLitWord8Wrap
-        , mkLitInt16, mkLitInt16Wrap
-        , mkLitWord16, mkLitWord16Wrap
-        , mkLitInt32, mkLitInt32Wrap
-        , mkLitWord32, mkLitWord32Wrap
-        , mkLitInt64, mkLitInt64Wrap
-        , mkLitWord64, mkLitWord64Wrap
+        , mkLitWord, mkLitWordWrap, mkLitWordWrapC, mkLitWordUnchecked
+        , mkLitInt8, mkLitInt8Wrap, mkLitInt8Unchecked
+        , mkLitWord8, mkLitWord8Wrap, mkLitWord8Unchecked
+        , mkLitInt16, mkLitInt16Wrap, mkLitInt16Unchecked
+        , mkLitWord16, mkLitWord16Wrap, mkLitWord16Unchecked
+        , mkLitInt32, mkLitInt32Wrap, mkLitInt32Unchecked
+        , mkLitWord32, mkLitWord32Wrap, mkLitWord32Unchecked
+        , mkLitInt64, mkLitInt64Wrap, mkLitInt64Unchecked
+        , mkLitWord64, mkLitWord64Wrap, mkLitWord64Unchecked
         , mkLitFloat, mkLitDouble
         , mkLitChar, mkLitString
         , mkLitInteger, mkLitNatural
@@ -36,7 +36,6 @@ module GHC.Types.Literal
 
         -- ** Operations on Literals
         , literalType
-        , absentLiteralOf
         , pprLiteral
         , litNumIsSigned
         , litNumCheckRange
@@ -53,16 +52,15 @@ module GHC.Types.Literal
         , isZeroLit, isOneLit
         , litFitsInChar
         , litValue, mapLitValue
-        , isLitValue_maybe
+        , isLitValue_maybe, isLitRubbish
 
         -- ** Coercions
         , narrowInt8Lit, narrowInt16Lit, narrowInt32Lit, narrowInt64Lit
         , narrowWord8Lit, narrowWord16Lit, narrowWord32Lit, narrowWord64Lit
-        , extendIntLit, extendWordLit
+        , convertToIntLit, convertToWordLit
         , charToIntLit, intToCharLit
         , floatToIntLit, intToFloatLit, doubleToIntLit, intToDoubleLit
         , nullAddrLit, floatToDoubleLit, doubleToFloatLit
-        , rubbishLit, isRubbishLit
         ) where
 
 #include "HsVersions.h"
@@ -71,16 +69,13 @@ import GHC.Prelude
 
 import GHC.Builtin.Types.Prim
 import {-# SOURCE #-} GHC.Builtin.Types
-import GHC.Builtin.Names
 import GHC.Core.Type
-import GHC.Core.TyCon
 import GHC.Utils.Outputable
 import GHC.Data.FastString
 import GHC.Types.Basic
 import GHC.Utils.Binary
 import GHC.Settings.Constants
 import GHC.Platform
-import GHC.Types.Unique.FM
 import GHC.Utils.Misc
 import GHC.Utils.Panic
 
@@ -115,8 +110,7 @@ import Numeric ( fromRat )
 -- * The literal derived from the label mentioned in a \"foreign label\"
 --   declaration ('LitLabel')
 --
--- * A 'LitRubbish' to be used in place of values of 'UnliftedRep'
---   (i.e. 'MutVar#') when the value is never used.
+-- * A 'LitRubbish' to be used in place of values that are never used.
 --
 -- * A character
 -- * A string
@@ -139,10 +133,15 @@ data Literal
                                 -- that can be represented as a Literal. Create
                                 -- with 'nullAddrLit'
 
-  | LitRubbish Bool             -- ^ A nonsense value; always boxed, but
-                                --      True <=> lifted, False <=> unlifted
-                                -- Used when a binding is absent.
-                                -- See Note [Rubbish literals]
+  | LitRubbish Type             -- ^ A nonsense value of the given
+                                -- representation. See Note [Rubbish literals].
+                                --
+                                -- The Type argument, rr, is of kind RuntimeRep.
+                                -- The type of the literal is forall (a:TYPE rr). a
+                                --
+                                -- INVARIANT: the Type has no free variables
+                                --    and so substitution etc can ignore it
+                                --
 
   | LitFloat   Rational         -- ^ @Float#@. Create with 'mkLitFloat'
   | LitDouble  Rational         -- ^ @Double#@. Create with 'mkLitDouble'
@@ -220,7 +219,6 @@ instance Binary LitNumType where
 {-
 Note [BigNum literals]
 ~~~~~~~~~~~~~~~~~~~~~~
-
 GHC supports 2 kinds of arbitrary precision integers (a.k.a BigNum):
 
    * Natural: natural represented as a Word# or as a BigNat
@@ -234,7 +232,6 @@ are replaced with expression to build them at runtime from machine literals
 
 Note [String literals]
 ~~~~~~~~~~~~~~~~~~~~~~
-
 String literals are UTF-8 encoded and stored into ByteStrings in the following
 ASTs: Haskell, Core, Stg, Cmm. TH can also emit ByteString based string literals
 with the BytesPrimL constructor (see #14741).
@@ -263,7 +260,9 @@ instance Binary Literal where
         = do putByte bh 6
              put_ bh nt
              put_ bh i
-    put_ bh (LitRubbish b) = do putByte bh 7; put_ bh b
+    put_ _ (LitRubbish b) = pprPanic "Binary LitRubbish" (ppr b)
+     -- We use IfaceLitRubbish; see Note [Rubbish literals], item (6)
+
     get bh = do
             h <- getByte bh
             case h of
@@ -289,9 +288,7 @@ instance Binary Literal where
                     nt <- get bh
                     i  <- get bh
                     return (LitNumber nt i)
-              _ -> do
-                    b <- get bh
-                    return (LitRubbish b)
+              _ -> pprPanic "Binary:Literal" (int (fromIntegral h))
 
 instance Outputable Literal where
     ppr = pprLiteral id
@@ -572,6 +569,10 @@ mkLitNatural :: Integer -> Literal
 mkLitNatural x = ASSERT2( inNaturalRange x,  integer x )
                     (LitNumber LitNumNatural x)
 
+isLitRubbish :: Literal -> Bool
+isLitRubbish (LitRubbish {}) = True
+isLitRubbish _               = False
+
 inNaturalRange :: Integer -> Bool
 inNaturalRange x = x >= 0
 
@@ -682,13 +683,14 @@ narrowWord16Lit = narrowLit' @Word16 LitNumWord16
 narrowWord32Lit = narrowLit' @Word32 LitNumWord32
 narrowWord64Lit = narrowLit' @Word64 LitNumWord64
 
--- | Extend a fixed-width literal (e.g. 'Int16#') to a word-sized literal (e.g.
--- 'Int#').
-extendWordLit, extendIntLit :: Platform -> Literal -> Literal
-extendWordLit platform (LitNumber _nt i)  = mkLitWord platform i
-extendWordLit _platform l                 = pprPanic "extendWordLit" (ppr l)
-extendIntLit  platform (LitNumber _nt i)  = mkLitInt platform i
-extendIntLit  _platform l                 = pprPanic "extendIntLit" (ppr l)
+-- | Extend or narrow a fixed-width literal (e.g. 'Int16#') to a target
+-- word-sized literal ('Int#' or 'Word#'). Narrowing can only happen on 32-bit
+-- architectures when we convert a 64-bit literal into a 32-bit one.
+convertToWordLit, convertToIntLit :: Platform -> Literal -> Literal
+convertToWordLit platform (LitNumber _nt i)  = mkLitWordWrap platform i
+convertToWordLit _platform l                 = pprPanic "convertToWordLit" (ppr l)
+convertToIntLit  platform (LitNumber _nt i)  = mkLitIntWrap platform i
+convertToIntLit  _platform l                 = pprPanic "convertToIntLit" (ppr l)
 
 charToIntLit (LitChar c)       = mkLitIntUnchecked (toInteger (ord c))
 charToIntLit l                 = pprPanic "charToIntLit" (ppr l)
@@ -713,14 +715,6 @@ doubleToFloatLit l             = pprPanic "doubleToFloatLit" (ppr l)
 nullAddrLit :: Literal
 nullAddrLit = LitNullAddr
 
--- | A rubbish literal; see Note [Rubbish literals]
-rubbishLit :: Bool -> Literal
-rubbishLit is_lifted = LitRubbish is_lifted
-
-isRubbishLit :: Literal -> Bool
-isRubbishLit (LitRubbish {}) = True
-isRubbishLit _               = False
-
 {-
         Predicates
         ~~~~~~~~~~
@@ -816,7 +810,8 @@ litIsLifted (LitNumber nt _) = case nt of
   LitNumWord16  -> False
   LitNumWord32  -> False
   LitNumWord64  -> False
-litIsLifted _                  = False
+litIsLifted _                        = False
+  -- Even RUBBISH[LiftedRep] is unlifted, as rubbish values are always evaluated.
 
 {-
         Types
@@ -844,40 +839,12 @@ literalType (LitNumber lt _)  = case lt of
    LitNumWord16  -> word16PrimTy
    LitNumWord32  -> word32PrimTy
    LitNumWord64  -> word64PrimTy
-literalType (LitRubbish is_lifted) = mkForAllTy a Inferred (mkTyVarTy a)
+
+-- LitRubbish: see Note [Rubbish literals]
+literalType (LitRubbish rep)
+  = mkForAllTy a Inferred (mkTyVarTy a)
   where
-    -- See Note [Rubbish literals]
-    a | is_lifted = alphaTyVar
-      | otherwise = alphaTyVarUnliftedRep
-
-absentLiteralOf :: TyCon -> Maybe Literal
--- Return a literal of the appropriate primitive
--- TyCon, to use as a placeholder when it doesn't matter
--- Rubbish literals are handled in GHC.Core.Opt.WorkWrap.Utils, because
---  1. Looking at the TyCon is not enough, we need the actual type
---  2. This would need to return a type application to a literal
-absentLiteralOf tc = lookupUFM absent_lits tc
-
--- We do not use TyConEnv here to avoid import cycles.
-absent_lits :: UniqFM TyCon Literal
-absent_lits = listToUFM_Directly
-                        -- Explicitly construct the mape from the known
-                        -- keys of these tyCons.
-                        [ (addrPrimTyConKey,    LitNullAddr)
-                        , (charPrimTyConKey,    LitChar 'x')
-                        , (intPrimTyConKey,     mkLitIntUnchecked 0)
-                        , (int8PrimTyConKey,    mkLitInt8Unchecked 0)
-                        , (int16PrimTyConKey,   mkLitInt16Unchecked 0)
-                        , (int32PrimTyConKey,   mkLitInt32Unchecked 0)
-                        , (int64PrimTyConKey,   mkLitInt64Unchecked 0)
-                        , (wordPrimTyConKey,    mkLitWordUnchecked 0)
-                        , (word8PrimTyConKey,   mkLitWord8Unchecked 0)
-                        , (word16PrimTyConKey,  mkLitWord16Unchecked 0)
-                        , (word32PrimTyConKey,  mkLitWord32Unchecked 0)
-                        , (word64PrimTyConKey,  mkLitWord64Unchecked 0)
-                        , (floatPrimTyConKey,   LitFloat 0)
-                        , (doublePrimTyConKey,  LitDouble 0)
-                        ]
+    a = mkTemplateKindVar (tYPE rep)
 
 {-
         Comparison
@@ -893,7 +860,7 @@ cmpLit (LitDouble    a)     (LitDouble     b)     = a `compare` b
 cmpLit (LitLabel     a _ _) (LitLabel      b _ _) = a `lexicalCompareFS` b
 cmpLit (LitNumber nt1 a)    (LitNumber nt2  b)
   = (nt1 `compare` nt2) `mappend` (a `compare` b)
-cmpLit (LitRubbish b1)      (LitRubbish b2)       = b1 `compare` b2
+cmpLit (LitRubbish b1)      (LitRubbish b2)       = b1 `nonDetCmpType` b2
 cmpLit lit1 lit2
   | isTrue# (dataToTag# lit1 <# dataToTag# lit2) = LT
   | otherwise                                    = GT
@@ -929,9 +896,8 @@ pprLiteral add_par (LitLabel l mb fod) =
     where b = case mb of
               Nothing -> pprHsString l
               Just x  -> doubleQuotes (text (unpackFS l ++ '@':show x))
-pprLiteral _       (LitRubbish is_lifted)
-  = text "__RUBBISH"
-    <> parens (if is_lifted then text "lifted" else text "unlifted")
+pprLiteral _       (LitRubbish rep)
+  = text "RUBBISH" <> parens (ppr rep)
 
 pprIntegerVal :: (SDoc -> SDoc) -> Integer -> SDoc
 -- See Note [Printing of literals in Core].
@@ -973,61 +939,161 @@ LitFloat        -1.0#
 LitDouble       -1.0##
 LitInteger      -1                 (-1)
 LitLabel        "__label" ...      ("__label" ...)
-LitRubbish      "__RUBBISH"
+LitRubbish      "RUBBISH[...]"
 
 Note [Rubbish literals]
 ~~~~~~~~~~~~~~~~~~~~~~~
-During worker/wrapper after demand analysis, where an argument
-is unused (absent) we do the following w/w split (supposing that
-y is absent):
+Sometimes, we need to cough up a rubbish value of a certain type that is used
+in place of dead code we thus aim to eliminate. The value of a dead occurrence
+has no effect on the dynamic semantics of the program, so we can pick any value
+of the same representation.
 
-  f x y z = e
-===>
-  f x y z = $wf x z
-  $wf x z = let y = <absent value>
-            in e
+Exploiting the results of absence analysis in worker/wrapper is a scenario where
+we need such a rubbish value, see examples in Note [Absent fillers] in
+GHC.Core.Opt.WorkWrap.Utils.
 
-Usually the binding for y is ultimately optimised away, and
-even if not it should never be evaluated -- but that's the
-way the w/w split starts off.
+It's completely undefined what the *value* of a rubbish value is, e.g., we could
+pick @0#@ for @Int#@ or @42#@; it mustn't matter where it's inserted into a Core
+program. We embed these rubbish values in the 'LitRubbish' case of the 'Literal'
+data type. Here are the moving parts:
 
-What is <absent value>?
-* For lifted values <absent value> can be a call to 'error'.
-* For primitive types like Int# or Word# we can use any random
-  value of that type.
-* But what about /unlifted/ but /boxed/ types like MutVar# or
-  Array#?  Or /lifted/ but /strict/ values, such as a field of
-  a strict data constructor.  For these we use LitRubbish.
-  See Note [Absent errors] in GHC.Core.Opt.WorkWrap.Utils.hs
+1. Source Haskell: No way to produce rubbish lits in source syntax. Purely
+   an IR feature.
 
-The literal (LitRubbish is_lifted)
-has type
-  LitRubbish :: forall (a :: TYPE LiftedRep). a     if is_lifted
-  LitRubbish :: forall (a :: TYPE UnliftedRep). a   otherwise
+2. Core: 'LitRubbish' carries a `Type` of kind RuntimeRep,
+   describing the runtime representaion of the literal (is it a
+   pointer, an unboxed Double#, or whatever).
 
-So we might see a w/w split like
-  $wf x z = let y :: Array# Int = (LitRubbish False) @(Array# Int)
-            in e
+   We have it that `RUBBISH[rr]` has type `forall (a :: TYPE rr). a`.
+   See the `LitRubbish` case of `literalType`.
 
-Here are the moving parts, but see also Note [Absent errors] in
-GHC.Core.Opt.WorkWrap.Utils
+   The function GHC.Core.Make.mkLitRubbish makes a Core rubbish literal of
+   a given type.  It obeys the following invariants:
 
-* We define LitRubbish as a constructor in GHC.Types.Literal.Literal
+   INVARIANT 1: 'rr' has no free variables. Main reason: we don't need to run
+   substitutions and free variable finders over Literal. The rules around
+   levity/runtime-rep polymorphism naturally uphold this invariant.
 
-* It is given its polymorphic type by Literal.literalType
+   INVARIANT 2: we never make a rubbish literal of type (a ~# b). Reason:
+   see Note [Core type and coercion invariant] in GHC.Core.  We can't substitute
+   a LitRubbish inside a coercion, so it's best not to make one. They are zero
+   width anyway, so passing absent ones around costs nothing.  If we wanted
+   an absent filler of type (a ~# b) we should use (Coercion (UnivCo ...)),
+   but it doesn't seem worth making a new UnivCoProvenance for this purpose.
 
-* GHC.Core.Opt.WorkWrap.Utils.mk_absent_let introduces a LitRubbish for absent
-  arguments of boxed, unlifted type; or boxed, lifted arguments of strict data
-  constructors.
+   This is sad, though: see #18983.
 
-* In CoreToSTG we convert (RubishLit @t) to just ().  STG is untyped, so this
-  will work OK for both lifted and unlifted (but boxed) values. The important
-  thing is that it is a heap pointer, which the garbage collector can follow if
-  it encounters it.
+3. STG: The type app in `RUBBISH[IntRep] @Int# :: Int#` is erased and we get
+   the (untyped) 'StgLit' `RUBBISH[IntRep] :: Int#` in STG.
 
-  We considered maintaining LitRubbish in STG, and lowering it in the code
-  generators, but it seems simpler to do it once and for all in CoreToSTG.
+   It's treated mostly opaque, with the exception of the Unariser, where we
+   take apart a case scrutinisation on, or arg occurrence of, e.g.,
+   `RUBBISH[TupleRep[IntRep,DoubleRep]]` (which may stand in for `(# Int#, Double# #)`)
+   into its sub-parts `RUBBISH[IntRep]` and `RUBBISH[DoubleRep]`, similar to
+   unboxed tuples. `RUBBISH[VoidRep]` is erased.
+   See 'unariseRubbish_maybe' and also Note [Post-unarisation invariants].
 
-  In GHC.ByteCode.Asm we just lower it as a 0 literal, because it's all boxed to
-  the host GC anyway.
+4. Cmm: We translate 'LitRubbish' to their actual rubbish value in 'cgLit'.
+   The particulars are boring, and only matter when debugging illicit use of
+   a rubbish value; see Modes of failure below.
+
+5. Bytecode: In GHC.ByteCode.Asm we just lower it as a 0 literal, because it's
+   all boxed to the host GC anyway.
+
+6. IfaceSyn: `Literal` is part of `IfaceSyn`, but `Type` really isn't.  So in
+   the passage from Core to Iface I put LitRubbish into its owns IfaceExpr data
+   constructor, IfaceLitRubbish. The remaining constructors of Literal are
+   fine as IfaceSyn.
+
+Wrinkles
+
+a) Why do we put the `Type` (of kind RuntimeRep) inside the literal?  Could
+   we not instead /apply/ the literal to that RuntimeRep?  Alas no, becuase
+   then LitRubbish :: forall (rr::RuntimeRep) (a::TYPE rr). a
+   and that's am ill-formed type because its kind is `TYPE rr`, which escapes
+   the binding site of `rr`. Annoying.
+
+b) A rubbish literal is not bottom, and replies True to exprOkForSpeculation.
+   For unboxed types there is no bottom anyway.  If we have
+       let (x::Int#) = RUBBISH[IntRep] @Int#
+   we want to convert that to a case!  We want to leave it as a let, and
+   probably discard it as dead code soon after because x is unused.
+
+c) We can see a rubbish literal at the head of an application chain.
+   Most obviously, pretty much every rubbish literal is the head of a
+   type application e.g. `RUBBISH[IntRep] @Int#`.  But see also
+   Note [How a rubbish literal can be the head of an application]
+
+c) Literal is in Ord, because (and only because) we use Ord on AltCon when
+   building a TypeMap. Annoying.  We use `nonDetCmpType` here; the
+   non-determinism won't matter because it's only used in TrieMap.
+   Moreover, rubbish literals should not appear in patterns anyway.
+
+d) Why not lower LitRubbish in CoreToStg? Because it enables us to use
+   RubbishLit when unarising unboxed sums in the future, and it allows
+   rubbish values of e.g.  VecRep, for which we can't cough up dummy
+   values in STG.
+
+Modes of failure
+----------------
+Suppose there is a bug in GHC, and a rubbish value is used after all. That is
+undefined behavior, of course, but let us list a few examples for failure modes:
+
+ a) For an value of unboxed numeric type like `Int#`, we just use a silly
+    value like 42#. The error might propoagate indefinitely, hence we better
+    pick a rather unique literal. Same for Word, Floats, Char and VecRep.
+ b) For AddrRep (like String lits), we mit a null pointer, resulting in a
+    definitive segfault when accessed.
+ c) For boxed values, unlifted or not, we use a pointer to a fixed closure,
+    like `()`, so that the GC has a pointer to follow.
+    If we use that pointer as an 'Array#', we will likely access fields of the
+    array that don't exist, and a seg-fault is likely, but not guaranteed.
+    If we use that pointer as `Either Int Bool`, we might try to access the
+    'Int' field of the 'Left' constructor (which has the same ConTag as '()'),
+    which doesn't exists. In the best case, we'll find an invalid pointer in its
+    position and get a seg-fault, in the worst case the error manifests only one
+    or two indirections later.
+
+Note [How a rubbish literal can be the head of an application]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Consider this (#19824):
+
+    h :: T3 -> Int -> blah
+    h _ (I# n) = ...
+
+    f :: (T1 -> T2 -> T3) -> T4 -> blah
+    f g x = ....(h (g n s) x)...
+
+Demand analysis finds that h does not use its first argument, and w/w's h to
+
+    {-# INLINE h #-}
+    h a b = case b of I# n -> $wh n
+
+Demand analysis also finds that f does not use its first arg,
+so the worker for f look like
+
+    $wf x = let g = RUBBISH in
+            ....(h (g n s) x)...
+
+Now we inline g to get:
+
+    $wf x = ....(h (RUBBISH n s) x)...
+
+And lo, until we inline `h`, we have that application of
+RUBBISH in $wf's RHS.  But surely `h` will inline? Not if the
+arguments look boring.  Well, RUBBISH doesn't look boring.  But it
+could be a bit more complicated like
+   f g x = let t = ...(g n s)...
+           in ...(h t x)...
+
+and now the call looks more boring.  Anyway, the point is that we
+might reasonably see RUBBISH at the head of an application chain.
+
+It would be fine to rewrite
+  RUBBISH @(ta->tb->tr) a b  --->   RUBBISH @tr
+but we don't currently do so.
+
+It is NOT ok to discard the entire continuation:
+  case RUBBISH @ty of DEFAULT -> blah
+does not return RUBBISH!
 -}
