@@ -1,9 +1,9 @@
 {-# LANGUAGE InstanceSigs, TypeOperators #-}
 module Builder (
     -- * Data types
-    ArMode (..), CcMode (..), ConfigurationInfo (..), GhcMode (..),
-    GhcPkgMode (..), HaddockMode (..), SphinxMode (..), TarMode (..),
-    Builder (..),
+    ArMode (..), CcMode (..), ConfigurationInfo (..), DependencyType (..),
+    GhcMode (..), GhcPkgMode (..), HaddockMode (..), TestMode(..), SphinxMode (..),
+    TarMode (..), GitMode (..), Builder (..), Win32TarballsMode(..),
 
     -- * Builder properties
     builderProvenance, systemBuilderPath, builderPath, isSpecified, needBuilder,
@@ -38,15 +38,24 @@ import Oracles.Flag
 import Oracles.Setting (setting, Setting(..))
 import Packages
 
+import GHC.IO.Encoding (getFileSystemEncoding)
+import qualified Data.ByteString as BS
+import qualified GHC.Foreign as GHC
+
 -- | C compiler can be used in two different modes:
 -- * Compile or preprocess a source file.
 -- * Extract source dependencies by passing @-MM@ command line argument.
-data CcMode = CompileC | FindCDependencies deriving (Eq, Generic, Show)
+data CcMode = CompileC | FindCDependencies DependencyType deriving (Eq, Generic, Show)
+data DependencyType = CDep | CxxDep deriving (Eq, Generic, Show)
 
 instance Binary   CcMode
 instance Hashable CcMode
 instance NFData   CcMode
 
+instance Binary   DependencyType
+instance Hashable DependencyType
+instance NFData   DependencyType
+
 -- | GHC can be used in four different modes:
 -- * Compile a Haskell source file.
 -- * Compile a C source file.
@@ -87,10 +96,10 @@ instance Hashable ConfigurationInfo
 instance NFData   ConfigurationInfo
 
 -- TODO: Do we really need all these modes? Why do we need 'Dependencies'? We
--- can extract dependencies using the Cabal library. Note: we used to also have
--- the @Init@ mode for initialising a new package database but we've deleted it.
+-- can extract dependencies using the Cabal library.
 -- | 'GhcPkg' can initialise a package database and register packages in it.
-data GhcPkgMode = Copy         -- ^ Copy a package from one database to another.
+data GhcPkgMode = Init         -- ^ Initialise an empty package database
+                | Copy         -- ^ Copy a package from one database to another.
                 | Dependencies -- ^ Compute package dependencies.
                 | Unregister   -- ^ Unregister a package.
                 | Update       -- ^ Update a package.
@@ -109,6 +118,28 @@ instance Binary   HaddockMode
 instance Hashable HaddockMode
 instance NFData   HaddockMode
 
+-- | The testsuite driver can be called in two different modes:
+-- * Actually run the tests
+-- * Get all the hadrian targets needed for the current test configuration
+data TestMode = RunTest | GetExtraDeps deriving (Eq, Generic, Show)
+
+instance Binary   TestMode
+instance Hashable TestMode
+instance NFData   TestMode
+
+-- | Git is used to create source distributions
+data GitMode = ListFiles deriving (Eq, Generic, Show)
+
+instance Binary   GitMode
+instance Hashable GitMode
+instance NFData   GitMode
+
+data Win32TarballsMode = ListTarballs | VerifyTarballs | DownloadTarballs deriving (Eq, Generic, Show)
+
+instance Binary   Win32TarballsMode
+instance Hashable Win32TarballsMode
+instance NFData   Win32TarballsMode
+
 -- | A 'Builder' is a (usually external) command invoked in a separate process
 -- via 'cmd'. Here are some examples:
 -- * 'Alex' is a lexical analyser generator that builds @Lexer.hs@ from @Lexer.x@.
@@ -143,12 +174,14 @@ data Builder = Alex
              | Patch
              | Python
              | Ranlib
-             | RunTest
+             | Testsuite TestMode
              | Sphinx SphinxMode
              | Tar TarMode
              | Unlit
              | Xelatex
              | Makeindex  -- ^ from xelatex
+             | Git GitMode
+             | Win32Tarballs Win32TarballsMode
              deriving (Eq, Generic, Show)
 
 instance Binary   Builder
@@ -220,7 +253,8 @@ instance H.Builder Builder where
     -- contrast this with runBuilderWith, which returns @Action ()@
     -- this returns the @stdout@ from running the builder.
     -- For now this only implements asking @ghc-pkg@ about package
-    -- dependencies.
+    -- dependencies and asking the testsuite driver about hadrian
+    -- dependencies for tests.
     askBuilderWith :: Builder -> BuildInfo -> Action String
     askBuilderWith builder BuildInfo {..} = case builder of
         GhcPkg Dependencies _ -> do
@@ -236,6 +270,26 @@ instance H.Builder Builder where
                 need [path]
             Stdout stdout <- cmd' [path] ["--no-user-package-db", "field", input, "depends"]
             return stdout
+        Testsuite GetExtraDeps -> do
+          path <- builderPath builder
+          withResources buildResources $ do
+              -- The testsuite driver reports the dependencies on stderr
+              -- buildArgs should include --only-report-hadrian-deps at this point
+              Stderr stderr <- cmd' [path] buildArgs
+              return stderr
+        Git ListFiles -> do
+          path <- builderPath builder
+          withResources buildResources $ do
+              -- NUL separated list of files
+              -- We need to read this in the filesystem encoding
+              enc <- liftIO getFileSystemEncoding
+              Stdout stdout <- cmd' BinaryPipes [path] buildArgs
+              liftIO $ BS.useAsCStringLen stdout $ \fp -> GHC.peekCStringLen enc fp
+        Win32Tarballs ListTarballs -> do
+          path <- builderPath builder
+          withResources buildResources $ do
+              Stdout stdout <- cmd' [path] buildArgs
+              pure stdout
         _ -> error $ "Builder " ++ show builder ++ " can not be asked!"
 
     runBuilderWith :: Builder -> BuildInfo -> Action ()
@@ -317,7 +371,10 @@ instance H.Builder Builder where
 
                 -- RunTest produces a very large amount of (colorised) output;
                 -- Don't attempt to capture it.
-                RunTest -> cmd echo [path] buildArgs
+                Testsuite RunTest -> do
+                  Exit code <- cmd echo [path] buildArgs
+                  when (code /= ExitSuccess) $ do
+                    fail "tests failed"
 
                 _  -> cmd' echo [path] buildArgs
 
@@ -363,11 +420,13 @@ systemBuilderPath builder = case builder of
     Patch           -> fromKey "patch"
     Python          -> fromKey "python"
     Ranlib          -> fromKey "ranlib"
-    RunTest         -> fromKey "python"
+    Testsuite _     -> fromKey "python"
     Sphinx _        -> fromKey "sphinx-build"
     Tar _           -> fromKey "tar"
+    Git _           -> fromKey "git"
     Xelatex         -> fromKey "xelatex"
     Makeindex       -> fromKey "makeindex"
+    Win32Tarballs _ -> fromKey "python"
     _               -> error $ "No entry for " ++ show builder ++ inCfg
   where
     inCfg = " in " ++ quote configFile ++ " file."
