@@ -32,12 +32,14 @@ import qualified Distribution.Simple.Program.Types             as C
 import qualified Distribution.Simple.Configure                 as C (getPersistBuildConfig)
 import qualified Distribution.Simple.Build                     as C
 import qualified Distribution.Types.ComponentLocalBuildInfo    as C
-import qualified Distribution.Types.ComponentRequestedSpec     as C
 import qualified Distribution.InstalledPackageInfo             as Installed
 import qualified Distribution.Simple.PackageIndex              as C
 import qualified Distribution.Text                             as C
 import qualified Distribution.Types.LocalBuildInfo             as C
 import qualified Distribution.Types.MungedPackageId            as C
+#if MIN_VERSION_Cabal(3,5,0)
+import qualified Distribution.Utils.Path                       as C
+#endif
 import qualified Distribution.Utils.ShortText                  as C
 #if !MIN_VERSION_Cabal(3,4,0)
 import qualified Distribution.Types.CondTree                   as C
@@ -191,43 +193,19 @@ registerPackage context@Context {..} = do
     traced "cabal-register" $
         C.defaultMainWithHooksNoReadArgs C.autoconfUserHooks gpd
             [ "register", "--builddir", ctxPath, v ]
+-- | What type of file is Main
+data MainSourceType = HsMain | CppMain | CMain
 
 -- | Parse the 'ContextData' of a given 'Context'.
 resolveContextData :: Context -> Action ContextData
 resolveContextData context@Context {..} = do
-    -- TODO: This is conceptually wrong!
-    -- We should use the gpd, the flagAssignment and compiler, hostPlatform, and
-    -- other information from the lbi. And then compute the finalised PD (flags,
-    -- satisfiable dependencies, platform, compiler info, deps, gpd).
-    --
-    -- let (Right (pd,_)) = C.finalizePackageDescription flags (const True) platform (compilerInfo compiler) [] gpd
-    --
-    -- However when using the new-build path's this might change.
-
-    -- Read the package description from the Cabal file
-    gpd <- genericPackageDescription <$> readPackageData package
-
-    -- Configure the package with the GHC for this stage
-    (compiler, platform) <- configurePackageGHC package stage
-
-    flagList <- interpret (target context (Cabal Flags stage) [] []) =<< args <$> flavour
-    let flags = foldr addFlag mempty flagList
-          where
-            addFlag :: String -> C.FlagAssignment -> C.FlagAssignment
-            addFlag ('-':name) = C.insertFlagAssignment (C.mkFlagName name) False
-            addFlag ('+':name) = C.insertFlagAssignment (C.mkFlagName name) True
-            addFlag name       = C.insertFlagAssignment (C.mkFlagName name) True
-
-    let (Right (pd,_)) = C.finalizePD flags C.defaultComponentRequestedSpec
-                         (const True) platform (C.compilerInfo compiler) [] gpd
-
     cPath <- Context.contextPath context
     lbi <- liftIO $ C.getPersistBuildConfig cPath
 
     -- Note: the @cPath@ is ignored. The path that's used is the 'buildDir' path
     -- from the local build info @lbi@.
     pdi <- liftIO $ getHookedBuildInfo [pkgPath package, cPath -/- "build"]
-    let pd'  = C.updatePackageDescription pdi pd
+    let pd'  = C.updatePackageDescription pdi (C.localPkgDescr lbi)
         lbi' = lbi { C.localPkgDescr = pd' }
 
     -- TODO: Get rid of deprecated 'externalPackageDeps' and drop -Wno-deprecations
@@ -265,15 +243,26 @@ resolveContextData context@Context {..} = do
             -- @library-dirs@ here.
             _ -> error "No (or multiple) GHC rts package is registered!"
 
-        (buildInfo, modules, mainIs) = biModules pd'
+        (buildInfo, modules, mainIs) = biModules (C.localPkgDescr lbi')
 
-      in return $ ContextData
+        classifyMain :: FilePath -> MainSourceType
+        classifyMain fp
+          | takeExtension fp `elem` [".hs", ".lhs"] = HsMain
+          | takeExtension fp `elem` [".cpp", ".cxx", ".c++"]= CppMain
+          | otherwise = CMain
+
+        main_src = fmap (first C.display) mainIs
+        cdata = ContextData
           { dependencies    = deps
           , componentId     = C.localCompatPackageKey lbi'
-          , mainIs          = fmap (first C.display) mainIs
+          , mainIs          = main_src
           , modules         = map C.display modules
           , otherModules    = map C.display $ C.otherModules buildInfo
-          , srcDirs         = C.hsSourceDirs buildInfo
+          , srcDirs         =
+#if MIN_VERSION_Cabal(3,5,0)
+                              map C.getSymbolicPath
+#endif
+                                  (C.hsSourceDirs buildInfo)
           , depIds          = depIds
           , depNames        = map (C.display . C.mungedName . snd) extDeps
           , includeDirs     = C.includeDirs     buildInfo
@@ -282,8 +271,8 @@ resolveContextData context@Context {..} = do
           , extraLibs       = C.extraLibs       buildInfo
           , extraLibDirs    = C.extraLibDirs    buildInfo
           , asmSrcs         = C.asmSources      buildInfo
-          , cSrcs           = C.cSources        buildInfo
-          , cxxSrcs         = C.cxxSources      buildInfo
+          , cSrcs           = C.cSources        buildInfo ++ [ ms | Just (_,ms) <- pure main_src, CMain   <- pure (classifyMain ms)]
+          , cxxSrcs         = C.cxxSources      buildInfo ++ [ ms | Just (_,ms) <- pure main_src, CppMain <- pure (classifyMain ms)]
           , cmmSrcs         = C.cmmSources      buildInfo
           , hcOpts          = C.programDefaultArgs ghcProg
               ++ C.hcOptions C.GHC buildInfo
@@ -303,6 +292,8 @@ resolveContextData context@Context {..} = do
           , frameworks         = C.frameworks buildInfo
           , packageDescription = pd' }
 
+      in return cdata
+
 -- | Build autogenerated files @autogen/cabal_macros.h@ and @autogen/Paths_*.hs@.
 buildAutogenFiles :: Context -> Action ()
 buildAutogenFiles context = do
