@@ -50,8 +50,9 @@ compilerDependencies = do
     ghcPath <- expr $ buildPath (vanillaContext stage compiler)
     rtsPath <- expr (rtsBuildPath stage)
     libDir <- expr $ stageLibPath stage
+    useSystemFfi <- expr (flag UseSystemFfi)
     mconcat [ return $ (libDir -/-) <$> derivedConstantsFiles
-            , notStage0 ? return ((rtsPath -/-) <$> libffiHeaderFiles)
+            , notStage0 ? not useSystemFfi ? return ((rtsPath -/-) <$> libffiHeaderFiles)
             , return $ fmap (ghcPath -/-)
                   [ "primop-can-fail.hs-incl"
                   , "primop-code-size.hs-incl"
@@ -76,11 +77,12 @@ generatedDependencies :: Expr [FilePath]
 generatedDependencies = do
     stage    <- getStage
     rtsPath  <- expr (rtsBuildPath stage)
+    useSystemFfi <- expr (flag UseSystemFfi)
     includes <- expr $ includesDependencies stage
     libDir <- expr $ stageLibPath stage
     mconcat [ package compiler ? compilerDependencies
             , package ghcPrim  ? ghcPrimDependencies
-            , package rts      ? return (fmap (rtsPath -/-) libffiHeaderFiles
+            , package rts      ? return ((if useSystemFfi then pure [] else fmap (rtsPath -/-) libffiHeaderFiles)
                 ++ includes
                 ++ ((libDir -/-) <$> derivedConstantsFiles))
             , stage0 ? return includes ]
@@ -334,7 +336,6 @@ generateSettings = do
         , ("LLVM opt command", expr $ settingsFileSetting SettingsFileSetting_OptCommand)
         , ("LLVM clang command", expr $ settingsFileSetting SettingsFileSetting_ClangCommand)
 
-        , ("BigNum backend", getBignumBackend)
         , ("Use interpreter", expr $ yesNo <$> ghcWithInterpreter)
         , ("Support SMP", expr $ yesNo <$> targetSupportsSMP)
         , ("RTS ways", unwords . map show <$> getRtsWays)
