@@ -7,9 +7,9 @@ import Distribution.Client.RebuildMonad
          ( execRebuild )
 import Distribution.Simple.Program
 import Distribution.Verbosity as Verbosity
-import Distribution.Client.SourceRepo (SourceRepositoryPackage (..), SourceRepoProxy)
+import Distribution.Client.Types.SourceRepo (SourceRepositoryPackage (..), SourceRepoProxy)
 
-import Data.List
+import Data.List (mapAccumL)
 import Data.Tuple
 import qualified Data.Map as Map
 import qualified Data.Set as Set
@@ -47,29 +47,26 @@ import UnitTests.TempTestDir (withTestDir, removeDirectoryRecursiveHack)
 --
 tests :: MTimeChange -> [TestTree]
 tests mtimeChange =
-  [ testGroup "check VCS test framework" $
-    [ testProperty "git"    prop_framework_git
-    ] ++
-    [ testProperty "darcs" (prop_framework_darcs mtimeChange)
-    | enableDarcsTests
+  [ testGroup "git"
+    [ testProperty "check VCS test framework"    prop_framework_git
+    , testProperty "cloneSourceRepo"             prop_cloneRepo_git
+    , testProperty "syncSourceRepos"             prop_syncRepos_git
     ]
-  , testGroup "cloneSourceRepo" $
-    [ testProperty "git"    prop_cloneRepo_git
-    ] ++
-    [ testProperty "darcs" (prop_cloneRepo_darcs mtimeChange)
-    | enableDarcsTests
-    ]
-  , testGroup "syncSourceRepos" $
-    [ testProperty "git"    prop_syncRepos_git
-    ] ++
-    [ testProperty "darcs" (prop_syncRepos_darcs mtimeChange)
-    | enableDarcsTests
-    ]
-  ]
-  where
-    -- for the moment they're not yet working
-    enableDarcsTests = False
 
+    -- for the moment they're not yet working
+  , testGroup "darcs" $ const []
+    [ testProperty "check VCS test framework"    $ prop_framework_darcs mtimeChange
+    , testProperty "cloneSourceRepo"             $ prop_cloneRepo_darcs mtimeChange
+    , testProperty "syncSourceRepos"             $ prop_syncRepos_darcs mtimeChange
+    ]
+
+  , testGroup "pijul" $ const []
+    [ testProperty "check VCS test framework"    prop_framework_pijul
+    , testProperty "cloneSourceRepo"             prop_cloneRepo_pijul
+    , testProperty "syncSourceRepos"             prop_syncRepos_pijul
+    ]
+
+  ]
 
 prop_framework_git :: BranchingRepoRecipe -> Property
 prop_framework_git =
@@ -83,6 +80,12 @@ prop_framework_darcs mtimeChange =
   . prop_framework vcsDarcs (vcsTestDriverDarcs mtimeChange)
   . WithoutBranchingSupport
 
+prop_framework_pijul :: BranchingRepoRecipe -> Property
+prop_framework_pijul =
+    ioProperty
+  . prop_framework vcsPijul vcsTestDriverPijul
+  . WithBranchingSupport
+
 prop_cloneRepo_git :: BranchingRepoRecipe -> Property
 prop_cloneRepo_git =
     ioProperty
@@ -96,6 +99,12 @@ prop_cloneRepo_darcs mtimeChange =
   . prop_cloneRepo vcsDarcs (vcsTestDriverDarcs mtimeChange)
   . WithoutBranchingSupport
 
+prop_cloneRepo_pijul :: BranchingRepoRecipe -> Property
+prop_cloneRepo_pijul =
+    ioProperty
+  . prop_cloneRepo vcsPijul vcsTestDriverPijul
+  . WithBranchingSupport
+
 prop_syncRepos_git :: RepoDirSet -> SyncTargetIterations -> PrngSeed
                    -> BranchingRepoRecipe -> Property
 prop_syncRepos_git destRepoDirs syncTargetSetIterations seed =
@@ -113,6 +122,13 @@ prop_syncRepos_darcs  mtimeChange destRepoDirs syncTargetSetIterations seed =
                    destRepoDirs syncTargetSetIterations seed
   . WithoutBranchingSupport
 
+prop_syncRepos_pijul :: RepoDirSet -> SyncTargetIterations -> PrngSeed
+                   -> BranchingRepoRecipe -> Property
+prop_syncRepos_pijul destRepoDirs syncTargetSetIterations seed =
+    ioProperty
+  . prop_syncRepos vcsPijul vcsTestDriverPijul
+                   destRepoDirs syncTargetSetIterations seed
+  . WithBranchingSupport
 
 -- ------------------------------------------------------------
 -- * General test setup
@@ -199,6 +215,7 @@ prop_cloneRepo vcs mkVCSTestDriver repoRecipe =
             , srpTag      = Just tagname
             , srpBranch   = Nothing
             , srpSubdir   = []
+            , srpCommand  = []
             }
     verbosity = silent
 
@@ -299,6 +316,7 @@ pickSyncTargetSets repoType repoState srcRepoPath dstReposPath =
                               , srpTag      = Just tag
                               , srpBranch   = Nothing
                               , srpSubdir   = Proxy
+                              , srpCommand  = []
                               }
         (tag, workingState) = Map.elemAt tagIdx (allTags repoState)
         (tagIdx, prng')     = randomR (0, Map.size (allTags repoState) - 1) prng
@@ -693,3 +711,47 @@ vcsTestDriverDarcs mtimeChange verbosity vcs repoRoot =
                            }
     darcs = runProgramInvocation verbosity . darcsInvocation
 
+
+vcsTestDriverPijul :: Verbosity -> VCS ConfiguredProgram
+                 -> FilePath -> VCSTestDriver
+vcsTestDriverPijul verbosity vcs repoRoot =
+    VCSTestDriver {
+      vcsVCS = vcs
+
+    , vcsRepoRoot = repoRoot
+
+    , vcsIgnoreFiles = Set.empty
+
+    , vcsInit =
+        pijul $ ["init"]
+
+    , vcsAddFile = \_ filename ->
+        pijul ["add", filename]
+
+    , vcsCommitChanges = \_state -> do
+        pijul $ ["record", "-a", "-m 'a patch'"
+                , "-A 'A <a@example.com>'"
+                ]
+        commit <- pijul' ["log"]
+        let commit' = takeWhile (not . isSpace) commit
+        return (Just commit')
+
+    -- tags work differently in pijul...
+    -- so this is wrong
+    , vcsTagState = \_ tagname ->
+        pijul ["tag", tagname]
+
+    , vcsSwitchBranch = \_ branchname -> do
+--        unless (branchname `Map.member` allBranches) $
+--          pijul ["from-branch", branchname]
+        pijul $ ["checkout", branchname]
+
+    , vcsCheckoutTag = Left $ \tagname ->
+        pijul $ ["checkout", tagname]
+    }
+  where
+    gitInvocation args = (programInvocation (vcsProgram vcs) args) {
+                           progInvokeCwd = Just repoRoot
+                         }
+    pijul  = runProgramInvocation       verbosity . gitInvocation
+    pijul' = getProgramInvocationOutput verbosity . gitInvocation
