@@ -1,17 +1,5 @@
 {-
 
-NOTA BENE: Do NOT use ($) anywhere in this module! The type of ($) is
-slightly magical (it can return unlifted types), and it is wired in.
-But, it is also *defined* in this module, with a non-magical type.
-GHC gets terribly confused (and *hangs*) if you try to use ($) in this
-module, because it has different types in different scenarios.
-
-This is not a problem in general, because the type ($), being wired in, is not
-written out to the interface file, so importing files don't get confused.
-The problem is only if ($) is used here. So don't!
-
----------------------------------------------
-
 The overall structure of the GHC Prelude is a bit tricky.
 
   a) We want to avoid "orphan modules", i.e. ones with instance
@@ -74,19 +62,17 @@ GHC.Float       Classes: Floating, RealFloat
 Other Prelude modules are much easier with fewer complex dependencies.
 -}
 
+{-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DataKinds #-}
+{-# LANGUAGE ExistentialQuantification #-}
+{-# LANGUAGE MagicHash #-}
+{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE PolyKinds #-}
+{-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE UnboxedTuples #-}
 {-# LANGUAGE Unsafe #-}
-{-# LANGUAGE CPP
-           , NoImplicitPrelude
-           , BangPatterns
-           , ExplicitForAll
-           , MagicHash
-           , UnboxedTuples
-           , ExistentialQuantification
-           , RankNTypes
-           , KindSignatures
-           , PolyKinds
-           , DataKinds
-  #-}
+
 -- -Wno-orphans is needed for things like:
 -- Orphan rule: "x# -# x#" ALWAYS forall x# :: Int# -# x# x# = 0
 {-# OPTIONS_GHC -Wno-orphans #-}
@@ -132,9 +118,8 @@ import GHC.Err
 import GHC.Maybe
 import {-# SOURCE #-} GHC.IO (mkUserError, mplusIO)
 
-import GHC.Tuple ()              -- Note [Depend on GHC.Tuple]
-import GHC.Integer ()            -- Note [Depend on GHC.Integer]
-import GHC.Natural ()            -- Note [Depend on GHC.Natural]
+import GHC.Tuple (Solo (..))     -- Note [Depend on GHC.Tuple]
+import GHC.Num.Integer ()        -- Note [Depend on GHC.Num.Integer]
 
 -- for 'class Semigroup'
 import {-# SOURCE #-} GHC.Real (Integral)
@@ -144,6 +129,9 @@ import {-# SOURCE #-} Data.Semigroup.Internal ( stimesDefault
                                               , stimesIdempotentMonoid
                                               )
 
+-- $setup
+-- >>> import GHC.Num
+
 infixr 9  .
 infixr 5  ++
 infixl 4  <$
@@ -156,40 +144,42 @@ infixl 4 <*>, <*, *>, <**>
 default ()              -- Double isn't available yet
 
 {-
-Note [Depend on GHC.Integer]
-~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-The Integer type is special because TidyPgm uses
-GHC.Integer.Type.mkInteger to construct Integer literal values
-Currently it reads the interface file whether or not the current
-module *has* any Integer literals, so it's important that
-GHC.Integer.Type (in package integer-gmp or integer-simple) is
-compiled before any other module.  (There's a hack in GHC to disable
-this for packages ghc-prim, integer-gmp, integer-simple, which aren't
-allowed to contain any Integer literals.)
+Note [Depend on GHC.Num.Integer]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-Likewise we implicitly need Integer when deriving things like Eq
-instances.
+The Integer type is special because GHC.CoreToStg.Prep.mkConvertNumLiteral
+lookups names in ghc-bignum interfaces to construct Integer literal values.
+Currently it reads the interface file whether or not the current module *has*
+any Integer literals, so it's important that GHC.Num.Integer is compiled before
+any other module.
 
-The danger is that if the build system doesn't know about the dependency
-on Integer, it'll compile some base module before GHC.Integer.Type,
+The danger is that if the build system doesn't know about the implicit
+dependency on Integer, it'll compile some base module before GHC.Num.Integer,
 resulting in:
-  Failed to load interface for ‘GHC.Integer.Type’
-    There are files missing in the ‘integer-gmp’ package,
+  Failed to load interface for ‘GHC.Num.Integer’
+    There are files missing in the ‘ghc-bignum’ package,
+
+Note that this is only a problem with the make-based build system. Hadrian
+doesn't interleave compilation of modules from separate packages and respects
+the dependency between `base` and `ghc-bignum`.
+
+To ensure that GHC.Num.Integer is there, we must ensure that there is a visible
+dependency on GHC.Num.Integer from every module in base.  We make GHC.Base
+depend on GHC.Num.Integer; and everything else either depends on GHC.Base,
+directly on GHC.Num.Integer, or does not have NoImplicitPrelude (and hence
+depends on Prelude).
+
+The lookup is only disabled for packages ghc-prim and ghc-bignum, which aren't
+allowed to contain any Integer literal.
 
-Bottom line: we make GHC.Base depend on GHC.Integer; and everything
-else either depends on GHC.Base, or does not have NoImplicitPrelude
-(and hence depends on Prelude).
 
 Note [Depend on GHC.Tuple]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~
 Similarly, tuple syntax (or ()) creates an implicit dependency on
 GHC.Tuple, so we use the same rule as for Integer --- see Note [Depend on
-GHC.Integer] --- to explain this to the build system.  We make GHC.Base
+GHC.Num.Integer] --- to explain this to the build system.  We make GHC.Base
 depend on GHC.Tuple, and everything else depends on GHC.Base or Prelude.
 
-Note [Depend on GHC.Natural]
-~~~~~~~~~~~~~~~~~~~~~~~~~~
-Similar to GHC.Integer.
 -}
 
 #if 0
@@ -231,7 +221,7 @@ class Semigroup a where
         -- The default definition should be sufficient, but this can be
         -- overridden for efficiency.
         --
-        -- >>> import Data.List.NonEmpty
+        -- >>> import Data.List.NonEmpty (NonEmpty (..))
         -- >>> sconcat $ "Hello" :| [" ", "Haskell", "!"]
         -- "Hello Haskell!"
         sconcat :: NonEmpty a -> a
@@ -300,8 +290,9 @@ class Semigroup a => Monoid a where
         -- >>> mconcat ["Hello", " ", "Haskell", "!"]
         -- "Hello Haskell!"
         mconcat :: [a] -> a
-        {-# INLINE mconcat #-}
         mconcat = foldr mappend mempty
+        {-# INLINE mconcat #-}
+        -- INLINE in the hope of fusion with mconcat's argument (see !4890)
 
 -- | @since 4.9.0.0
 instance Semigroup [a] where
@@ -332,7 +323,7 @@ mechanism to define mconcat and the Applicative and Monad instances for lists.
 We mark them INLINE because the inliner is not generally too keen to inline
 build forms such as the ones these desugar to without our insistence.  Defining
 these using list comprehensions instead of foldr has an additional potential
-benefit, as described in compiler/deSugar/DsListComp.hs: if optimizations
+benefit, as described in compiler/GHC/HsToCore/ListComp.hs: if optimizations
 needed to make foldr/build forms efficient are turned off, we'll get reasonably
 efficient translations anyway.
 -}
@@ -362,6 +353,15 @@ instance Monoid () where
         mempty        = ()
         mconcat _     = ()
 
+-- | @since 4.15
+instance Semigroup a => Semigroup (Solo a) where
+  Solo a <> Solo b = Solo (a <> b)
+  stimes n (Solo a) = Solo (stimes n a)
+
+-- | @since 4.15
+instance Monoid a => Monoid (Solo a) where
+  mempty = Solo mempty
+
 -- | @since 4.9.0.0
 instance (Semigroup a, Semigroup b) => Semigroup (a, b) where
         (a,b) <> (a',b') = (a<>a',b<>b')
@@ -436,6 +436,20 @@ instance Semigroup a => Semigroup (Maybe a) where
 instance Semigroup a => Monoid (Maybe a) where
     mempty = Nothing
 
+-- | @since 4.15
+instance Applicative Solo where
+  pure = Solo
+
+  -- Note: we really want to match strictly here. This lets us write,
+  -- for example,
+  --
+  -- forceSpine :: Foldable f => f a -> ()
+  -- forceSpine xs
+  --   | Solo r <- traverse_ Solo xs
+  --   = r
+  Solo f <*> Solo x = Solo (f x)
+  liftA2 f (Solo x) (Solo y) = Solo (f x y)
+
 -- | For tuples, the 'Monoid' constraint on @a@ determines
 -- how the first values merge.
 -- For example, 'String's concatenate:
@@ -449,6 +463,10 @@ instance Monoid a => Applicative ((,) a) where
     (u, f) <*> (v, x) = (u <> v, f x)
     liftA2 f (u, x) (v, y) = (u <> v, f x y)
 
+-- | @since 4.15
+instance Monad Solo where
+  Solo x >>= f = f x
+
 -- | @since 4.9.0.0
 instance Monoid a => Monad ((,) a) where
     (u, a) >>= k = case k a of (v, b) -> (u <> v, b)
@@ -498,31 +516,62 @@ Note, that the second law follows from the free theorem of the type 'fmap' and
 the first law, so you need only check that the former condition holds.
 -}
 
-class  Functor f  where
-    -- | Using @ApplicativeDo@: \'@'fmap' f as@\' can be understood as
-    -- the @do@ expression
+class Functor f where
+    -- | 'fmap' is used to apply a function of type @(a -> b)@ to a value of type @f a@,
+    -- where f is a functor, to produce a value of type @f b@.
+    -- Note that for any type constructor with more than one parameter (e.g., `Either`),
+    -- only the last type parameter can be modified with `fmap` (e.g., `b` in `Either a b`).
     --
-    -- @
-    -- do a <- as
-    --    pure (f a)
-    -- @
+    -- Some type constructors with two parameters or more have a @'Data.Bifunctor'@ instance that allows
+    -- both the last and the penultimate parameters to be mapped over.
     --
-    -- with an inferred @Functor@ constraint.
+    -- ==== __Examples__
+    --
+    -- Convert from a @'Data.Maybe.Maybe' Int@ to a @Maybe String@
+    -- using 'Prelude.show':
+    --
+    -- >>> fmap show Nothing
+    -- Nothing
+    -- >>> fmap show (Just 3)
+    -- Just "3"
+    --
+    -- Convert from an @'Data.Either.Either' Int Int@ to an
+    -- @Either Int String@ using 'Prelude.show':
+    --
+    -- >>> fmap show (Left 17)
+    -- Left 17
+    -- >>> fmap show (Right 17)
+    -- Right "17"
+    --
+    -- Double each element of a list:
+    --
+    -- >>> fmap (*2) [1,2,3]
+    -- [2,4,6]
+    --
+    -- Apply 'Prelude.even' to the second element of a pair:
+    --
+    -- >>> fmap even (2,2)
+    -- (2,True)
+    --
+    -- It may seem surprising that the function is only applied to the last element of the tuple
+    -- compared to the list example above which applies it to every element in the list.
+    -- To understand, remember that tuples are type constructors with multiple type parameters:
+    -- a tuple of 3 elements @(a,b,c)@ can also be written @(,,) a b c@ and its @Functor@ instance
+    -- is defined for @Functor ((,,) a b)@ (i.e., only the third parameter is free to be mapped over
+    -- with @fmap@).
+    --
+    -- It explains why @fmap@ can be used with tuples containing values of different types as in the
+    -- following example:
+    --
+    -- >>> fmap even ("hello", 1.0, 4)
+    -- ("hello",1.0,True)
+
     fmap        :: (a -> b) -> f a -> f b
 
     -- | Replace all locations in the input with the same value.
     -- The default definition is @'fmap' . 'const'@, but this may be
     -- overridden with a more efficient version.
     --
-    -- Using @ApplicativeDo@: \'@a '<$' bs@\' can be understood as the
-    -- @do@ expression
-    --
-    -- @
-    -- do bs
-    --    pure a
-    -- @
-    --
-    -- with an inferred @Functor@ constraint.
     (<$)        :: a -> f b -> f a
     (<$)        =  fmap . const
 
@@ -600,14 +649,18 @@ class Functor f => Applicative f where
     -- A few functors support an implementation of '<*>' that is more
     -- efficient than the default one.
     --
-    -- Using @ApplicativeDo@: \'@fs '<*>' as@\' can be understood as
-    -- the @do@ expression
+    -- ==== __Example__
+    -- Used in combination with @('<$>')@, @('<*>')@ can be used to build a record.
     --
-    -- @
-    -- do f <- fs
-    --    a <- as
-    --    pure (f a)
-    -- @
+    -- >>> data MyState = MyState {arg1 :: Foo, arg2 :: Bar, arg3 :: Baz}
+    --
+    -- >>> produceFoo :: Applicative f => f Foo
+    --
+    -- >>> produceBar :: Applicative f => f Bar
+    -- >>> produceBaz :: Applicative f => f Baz
+    --
+    -- >>> mkState :: Applicative f => f MyState
+    -- >>> mkState = MyState <$> produceFoo <*> produceBar <*> produceBaz
     (<*>) :: f (a -> b) -> f a -> f b
     (<*>) = liftA2 id
 
@@ -621,38 +674,38 @@ class Functor f => Applicative f where
     -- This became a typeclass method in 4.10.0.0. Prior to that, it was
     -- a function defined in terms of '<*>' and 'fmap'.
     --
-    -- Using @ApplicativeDo@: \'@'liftA2' f as bs@\' can be understood
-    -- as the @do@ expression
-    --
-    -- @
-    -- do a <- as
-    --    b <- bs
-    --    pure (f a b)
-    -- @
+    -- ==== __Example__
+    -- >>> liftA2 (,) (Just 3) (Just 5)
+    -- Just (3,5)
 
     liftA2 :: (a -> b -> c) -> f a -> f b -> f c
     liftA2 f x = (<*>) (fmap f x)
 
     -- | Sequence actions, discarding the value of the first argument.
     --
-    -- \'@as '*>' bs@\' can be understood as the @do@ expression
+    -- ==== __Examples__
+    -- If used in conjunction with the Applicative instance for 'Maybe',
+    -- you can chain Maybe computations, with a possible "early return"
+    -- in case of 'Nothing'.
     --
-    -- @
-    -- do as
-    --    bs
-    -- @
+    -- >>> Just 2 *> Just 3
+    -- Just 3
     --
-    -- This is a tad complicated for our @ApplicativeDo@ extension
-    -- which will give it a @Monad@ constraint. For an @Applicative@
-    -- constraint we write it of the form
+    -- >>> Nothing *> Just 3
+    -- Nothing
     --
-    -- @
-    -- do _ <- as
-    --    b <- bs
-    --    pure b
-    -- @
+    -- Of course a more interesting use case would be to have effectful
+    -- computations instead of just returning pure values.
+    --
+    -- >>> import Data.Char
+    -- >>> import Text.ParserCombinators.ReadP
+    -- >>> let p = string "my name is " *> munch1 isAlpha <* eof
+    -- >>> readP_to_S p "my name is Simon"
+    -- [("Simon","")]
+
     (*>) :: f a -> f b -> f b
     a1 *> a2 = (id <$ a1) <*> a2
+
     -- This is essentially the same as liftA2 (flip const), but if the
     -- Functor instance has an optimized (<$), it may be better to use
     -- that instead. Before liftA2 became a method, this definition
@@ -664,60 +717,40 @@ class Functor f => Applicative f where
 
     -- | Sequence actions, discarding the value of the second argument.
     --
-    -- Using @ApplicativeDo@: \'@as '<*' bs@\' can be understood as
-    -- the @do@ expression
-    --
-    -- @
-    -- do a <- as
-    --    bs
-    --    pure a
-    -- @
     (<*) :: f a -> f b -> f a
     (<*) = liftA2 const
 
 -- | A variant of '<*>' with the arguments reversed.
 --
--- Using @ApplicativeDo@: \'@as '<**>' fs@\' can be understood as the
--- @do@ expression
---
--- @
--- do a <- as
---    f <- fs
---    pure (f a)
--- @
 (<**>) :: Applicative f => f a -> f (a -> b) -> f b
 (<**>) = liftA2 (\a f -> f a)
 -- Don't use $ here, see the note at the top of the page
 
 -- | Lift a function to actions.
--- This function may be used as a value for `fmap` in a `Functor` instance.
+-- Equivalent to Functor's `fmap` but implemented using only `Applicative`'s methods:
+-- `liftA f a = pure f <*> a`
 --
--- | Using @ApplicativeDo@: \'@'liftA' f as@\' can be understood as the
--- @do@ expression
+-- As such this function may be used to implement a `Functor` instance from an `Applicative` one.
+
 --
+-- ==== __Examples__
+-- Using the Applicative instance for Lists:
 --
--- @
--- do a <- as
---    pure (f a)
--- @
+-- >>> liftA (+1) [1, 2]
+-- [2,3]
 --
--- with an inferred @Functor@ constraint, weaker than @Applicative@.
+-- Or the Applicative instance for 'Maybe'
+--
+-- >>> liftA (+1) (Just 3)
+-- Just 4
+
 liftA :: Applicative f => (a -> b) -> f a -> f b
 liftA f a = pure f <*> a
 -- Caution: since this may be used for `fmap`, we can't use the obvious
 -- definition of liftA = fmap.
 
 -- | Lift a ternary function to actions.
---
--- Using @ApplicativeDo@: \'@'liftA3' f as bs cs@\' can be understood
--- as the @do@ expression
---
--- @
--- do a <- as
---    b <- bs
---    c <- cs
---    pure (f a b c)
--- @
+
 liftA3 :: Applicative f => (a -> b -> c -> d) -> f a -> f b -> f c -> f d
 liftA3 f a b c = liftA2 f a b <*> c
 
@@ -980,6 +1013,15 @@ instance Applicative ((->) r) where
 instance Monad ((->) r) where
     f >>= k = \ r -> k (f r) r
 
+-- | @since 4.15
+instance Functor Solo where
+  fmap f (Solo a) = Solo (f a)
+
+  -- Being strict in the `Solo` argument here seems most consistent
+  -- with the concept behind `Solo`: always strict in the wrapper and lazy
+  -- in the contents.
+  x <$ Solo _ = Solo x
+
 -- | @since 2.01
 instance Functor ((,) a) where
     fmap f (x,y) = (x, f y)
@@ -1252,8 +1294,7 @@ augment g xs = g (:) xs
 -- > map f [x1, x2, ...] == [f x1, f x2, ...]
 --
 -- >>> map (+1) [1, 2, 3]
---- [2,3,4]
-
+-- [2,3,4]
 map :: (a -> b) -> [a] -> [b]
 {-# NOINLINE [0] map #-}
   -- We want the RULEs "map" and "map/coerce" to fire first.
@@ -1314,6 +1355,7 @@ The rules for map work like this.
 --   http://research.microsoft.com/en-us/um/people/simonpj/papers/ext-f/coercible.pdf
 
 {-# RULES "map/coerce" [1] map coerce = coerce #-}
+-- See Note [Getting the map/coerce RULE to work] in CoreOpt
 
 ----------------------------------------------
 --              append
@@ -1371,7 +1413,7 @@ eqString (c1:cs1) (c2:cs2) = c1 == c2 && cs1 `eqString` cs2
 eqString _        _        = False
 
 {-# RULES "eqString" (==) = eqString #-}
--- eqString also has a BuiltInRule in PrelRules.hs:
+-- eqString also has a BuiltInRule in GHC.Core.Opt.ConstantFold:
 --      eqString (unpackCString# (Lit s1)) (unpackCString# (Lit s2)) = s1==s2
 
 
@@ -1476,6 +1518,7 @@ f $ x =  f x
 -- the function with that value.
 
 ($!) :: forall r a (b :: TYPE r). (a -> b) -> a -> b
+{-# INLINE ($!) #-}
 f $! x = let !vx = x in f vx  -- see #2273
 
 -- | @'until' p f@ yields the result of applying @f@ until @p@ holds.
@@ -1558,8 +1601,13 @@ getTag x = dataToTag# x
 -- Definitions of the boxed PrimOps; these will be
 -- used in the case of partial applications, etc.
 
+-- See Note [INLINE division wrappers]
 {-# INLINE quotInt #-}
 {-# INLINE remInt #-}
+{-# INLINE divInt #-}
+{-# INLINE modInt #-}
+{-# INLINE quotRemInt #-}
+{-# INLINE divModInt #-}
 
 quotInt, remInt, divInt, modInt :: Int -> Int -> Int
 (I# x) `quotInt`  (I# y) = I# (x `quotInt#` y)
@@ -1587,6 +1635,17 @@ x# `divModInt#` y#
  | otherwise                                =
                                     x# `quotRemInt#` y#
 
+{- Note [INLINE division wrappers]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+The Int division functions such as 'quotRemInt' and 'divModInt' have
+been manually worker/wrappered, presumably because they construct
+*nested* products.
+We intend to preserve the exact worker/wrapper split, hence we mark
+the wrappers INLINE (#19267). That makes sure the optimiser doesn't
+accidentally inline the worker into the wrapper, undoing the manual
+split again.
+-}
+
 -- Wrappers for the shift operations.  The uncheckedShift# family are
 -- undefined when the amount being shifted by is greater than the size
 -- in bits of Int#, so these wrappers perform a check and return
@@ -1637,7 +1696,13 @@ a `iShiftRL#` b | isTrue# (b >=# WORD_SIZE_IN_BITS#) = 0#
 "unpack-list"  [1]  forall a   . unpackFoldrCString# a (:) [] = unpackCString# a
 "unpack-append"     forall a n . unpackFoldrCString# a (:) n  = unpackAppendCString# a n
 
--- There's a built-in rule (in PrelRules.hs) for
+"unpack-utf8"       [~1] forall a   . unpackCStringUtf8# a             = build (unpackFoldrCStringUtf8# a)
+"unpack-list-utf8"  [1]  forall a   . unpackFoldrCStringUtf8# a (:) [] = unpackCStringUtf8# a
+"unpack-append-utf8"     forall a n . unpackFoldrCStringUtf8# a (:) n  = unpackAppendCStringUtf8# a n
+
+-- There's a built-in rule (in GHC.Core.Op.ConstantFold) for
 --      unpackFoldr "foo" c (unpackFoldr "baz" c n)  =  unpackFoldr "foobaz" c n
 
+-- See also the Note [String literals in GHC] in CString.hs
+
   #-}
