@@ -8,7 +8,7 @@
  * $Date: 2004/09/03 15:28:19 $
  * ---------------------------------------------------------------------------*/
 
-#ifdef DEBUG
+#if defined(DEBUG)
 
 #include "PosixSource.h"
 #include "Rts.h"
@@ -33,7 +33,7 @@ disInstr ( StgBCO *bco, int pc )
 
    StgWord16*     instrs      = (StgWord16*)(bco->instrs->payload);
 
-   StgArrWords*   literal_arr = bco->literals;
+   StgArrBytes*   literal_arr = bco->literals;
    StgWord*       literals    = (StgWord*)(&literal_arr->payload[0]);
 
    StgMutArrPtrs* ptrs_arr    = bco->ptrs;
@@ -67,15 +67,20 @@ disInstr ( StgBCO *bco, int pc )
    switch (instr & 0xff) {
       case bci_BRK_FUN:
          debugBelch ("BRK_FUN  " );  printPtr( ptrs[instrs[pc]] );
-         debugBelch (" %d ", instrs[pc+1]); printPtr( ptrs[instrs[pc+2]] ); debugBelch("\n" );
-         pc += 3;
+         debugBelch (" %d ", instrs[pc+1]); printPtr( ptrs[instrs[pc+2]] );
+         CostCentre* cc = (CostCentre*)literals[instrs[pc+3]];
+         if (cc) {
+           debugBelch(" %s", cc->label);
+         }
+         debugBelch("\n");
+         pc += 4;
          break;
       case bci_SWIZZLE:
          debugBelch("SWIZZLE stkoff %d by %d\n",
                          instrs[pc], (signed int)instrs[pc+1]);
          pc += 2; break;
       case bci_CCALL:
-         debugBelch("CCALL    marshaller at 0x%" FMT_Word "\n",
+         debugBelch("CCALL    marshaller at 0x%" FMT_HexWord "\n",
                          literals[instrs[pc]] );
          pc += 1; break;
      case bci_STKCHECK:  {
@@ -93,11 +98,28 @@ disInstr ( StgBCO *bco, int pc )
          debugBelch("PUSH_LLL %d %d %d\n", instrs[pc], instrs[pc+1],
                                                             instrs[pc+2] );
          pc += 3; break;
+      case bci_PUSH8:
+         debugBelch("PUSH8    %d\n", instrs[pc] );
+         pc += 1; break;
+      case bci_PUSH16:
+         debugBelch("PUSH16   %d\n", instrs[pc] );
+         pc += 1; break;
+      case bci_PUSH32:
+         debugBelch("PUSH32   %d\n", instrs[pc] );
+         pc += 1; break;
+      case bci_PUSH8_W:
+         debugBelch("PUSH8_W  %d\n", instrs[pc] );
+         pc += 1; break;
+      case bci_PUSH16_W:
+         debugBelch("PUSH16_W %d\n", instrs[pc] );
+         pc += 1; break;
+      case bci_PUSH32_W:
+         debugBelch("PUSH32_W %d\n", instrs[pc] );
+         pc += 1; break;
       case bci_PUSH_G:
          debugBelch("PUSH_G   " ); printPtr( ptrs[instrs[pc]] );
          debugBelch("\n" );
          pc += 1; break;
-
       case bci_PUSH_ALTS:
          debugBelch("PUSH_ALTS  " ); printPtr( ptrs[instrs[pc]] );
          debugBelch("\n");
@@ -126,11 +148,41 @@ disInstr ( StgBCO *bco, int pc )
          debugBelch("PUSH_ALTS_V  " ); printPtr( ptrs[instrs[pc]] );
          debugBelch("\n");
          pc += 1; break;
-
+      case bci_PUSH_ALTS_T:
+         debugBelch("PUSH_ALTS_T  ");
+         printPtr( ptrs[instrs[pc]] );
+         debugBelch(" 0x%" FMT_HexWord " ", literals[instrs[pc+1]] );
+         printPtr( ptrs[instrs[pc+2]] );
+         debugBelch("\n");
+         pc += 3; break;
+      case bci_PUSH_PAD8:
+         debugBelch("PUSH_PAD8\n");
+         pc += 1; break;
+      case bci_PUSH_PAD16:
+         debugBelch("PUSH_PAD16\n");
+         pc += 1; break;
+      case bci_PUSH_PAD32:
+         debugBelch("PUSH_PAD32\n");
+         pc += 1; break;
+      case bci_PUSH_UBX8:
+         debugBelch(
+             "PUSH_UBX8 0x%" FMT_HexWord8 "\n",
+             (StgWord8) literals[instrs[pc]] );
+         pc += 1; break;
+      case bci_PUSH_UBX16:
+         debugBelch(
+             "PUSH_UBX16 0x%" FMT_HexWord16 "\n",
+             (StgWord16) literals[instrs[pc]] );
+         pc += 1; break;
+      case bci_PUSH_UBX32:
+         debugBelch(
+             "PUSH_UBX32 0x%" FMT_HexWord32 "\n",
+             (StgWord32) literals[instrs[pc]] );
+         pc += 1; break;
       case bci_PUSH_UBX:
          debugBelch("PUSH_UBX ");
          for (i = 0; i < instrs[pc+1]; i++)
-            debugBelch("0x%" FMT_Word " ", literals[i + instrs[pc]] );
+            debugBelch("0x%" FMT_HexWord " ", literals[i + instrs[pc]] );
          debugBelch("\n");
          pc += 2; break;
       case bci_PUSH_APPLY_N:
@@ -265,6 +317,9 @@ disInstr ( StgBCO *bco, int pc )
       case bci_RETURN_V:
          debugBelch("RETURN_V\n" );
          break;
+      case bci_RETURN_T:
+         debugBelch("RETURN_T\n ");
+         break;
 
       default:
          barf("disInstr: unknown opcode %u", (unsigned int) instr);
@@ -272,19 +327,13 @@ disInstr ( StgBCO *bco, int pc )
    return pc;
 }
 
-
-/* Something of a kludge .. how do we know where the end of the insn
-   array is, since it isn't recorded anywhere?  Answer: the first
-   short is the number of bytecodes which follow it.
-   See ByteCodeGen.linkBCO.insns_arr for construction ...
-*/
 void disassemble( StgBCO *bco )
 {
-   nat i, j;
-   StgWord16*     instrs    = (StgWord16*)(bco->instrs->payload);
-   StgMutArrPtrs* ptrs      = bco->ptrs;
-   nat            nbcs      = (int)(bco->instrs->bytes / sizeof(StgWord16));
-   nat            pc        = 1;
+   uint32_t i, j;
+   StgWord16*     instrs  = (StgWord16*)(bco->instrs->payload);
+   StgMutArrPtrs* ptrs    = bco->ptrs;
+   uint32_t       nbcs    = (uint32_t)(bco->instrs->bytes / sizeof(StgWord16));
+   uint32_t       pc      = 1;
 
    debugBelch("BCO\n" );
    pc = 0;
