@@ -239,6 +239,7 @@ void initRtsFlagsDefaults(void)
     RtsFlags.TraceFlags.ticky         = false;
     RtsFlags.TraceFlags.trace_output  = NULL;
     RtsFlags.TraceFlags.eventlogFlushTime = 0;
+    RtsFlags.TraceFlags.nullWriter = false;
 #endif
 
 #if defined(PROFILING)
@@ -396,6 +397,7 @@ usage_text[] = {
 #else /* PROFILING */
 "  -h       Heap residency profile (output file <program>.hp)",
 "  -hT      Produce a heap profile grouped by closure type",
+"  -po<file>  Override profiling output file name prefix (program name by default)",
 #endif /* PROFILING */
 
 "  -i<sec>  Time between heap profile samples (seconds, default: 0.1)",
@@ -959,6 +961,11 @@ error = true;
                       OPTION_UNSAFE;
                       RtsFlags.MiscFlags.generate_dump_file = true;
                   }
+                  else if (strequal("null-eventlog-writer",
+                               &rts_argv[arg][2])) {
+                      OPTION_UNSAFE;
+                      RtsFlags.TraceFlags.nullWriter = true;
+                  }
                   else if (strequal("machine-readable",
                                &rts_argv[arg][2])) {
                       OPTION_UNSAFE;
@@ -1365,6 +1372,21 @@ error = true;
               case 'P': /* detailed cost centre profiling (time/alloc) */
               case 'p': /* cost centre profiling (time/alloc) */
                 OPTION_SAFE;
+#if !defined(PROFILING)
+                switch (rts_argv[arg][2]) {
+                  case 'o':
+                      if (rts_argv[arg][3] == '\0') {
+                        errorBelch("flag -po expects an argument");
+                        error = true;
+                        break;
+                      }
+                      RtsFlags.CcFlags.outputFileNameStem = rts_argv[arg]+3;
+                      break;
+                  default:
+                      PROFILING_BUILD_ONLY();
+
+                } break;
+#else
                 PROFILING_BUILD_ONLY(
                 switch (rts_argv[arg][2]) {
                   case 'a':
@@ -1399,6 +1421,7 @@ error = true;
                     goto check_rest;
                 }
                 ) break;
+#endif /* PROFILING */
 
               case 'R':
                   OPTION_SAFE;
