@@ -138,22 +138,37 @@ createThread(Capability *cap, W_ size)
     return tso;
 }
 
+/* ---------------------------------------------------------------------------
+ * Equality on Thread ids.
+ *
+ * This is used from STG land in the implementation of the Eq instance
+ * for ThreadIds.
+ * ------------------------------------------------------------------------ */
+
+bool
+eq_thread(StgPtr tso1, StgPtr tso2)
+{
+  return tso1 == tso2;
+}
+
 /* ---------------------------------------------------------------------------
  * Comparing Thread ids.
  *
- * This is used from STG land in the implementation of the
- * instances of Eq/Ord for ThreadIds.
+ * This is used from STG land in the implementation of the Ord instance
+ * for ThreadIds.
  * ------------------------------------------------------------------------ */
 
 int
 cmp_thread(StgPtr tso1, StgPtr tso2)
 {
+  if (tso1 == tso2) return 0;
+
   StgThreadID id1 = ((StgTSO *)tso1)->id;
   StgThreadID id2 = ((StgTSO *)tso2)->id;
 
-  if (id1 < id2) return (-1);
-  if (id1 > id2) return 1;
-  return 0;
+  ASSERT(id1 != id2);
+
+  return id1 < id2 ? -1 : 1;
 }
 
 /* ---------------------------------------------------------------------------
@@ -161,7 +176,7 @@ cmp_thread(StgPtr tso1, StgPtr tso2)
  *
  * This is used in the implementation of Show for ThreadIds.
  * ------------------------------------------------------------------------ */
-int
+long
 rts_getThreadId(StgPtr tso)
 {
   return ((StgTSO *)tso)->id;
@@ -865,12 +880,14 @@ printThreadBlockage(StgTSO *tso)
     debugBelch("is blocked until %ld", (long)(tso->block_info.target));
     break;
 #endif
+    break;
   case BlockedOnMVar:
     debugBelch("is blocked on an MVar @ %p", tso->block_info.closure);
     break;
   case BlockedOnMVarRead:
     debugBelch("is blocked on atomic MVar read @ %p", tso->block_info.closure);
     break;
+    break;
   case BlockedOnBlackHole:
       debugBelch("is blocked on a black hole %p",
                  ((StgBlockingQueue*)tso->block_info.bh->bh));
@@ -894,8 +911,8 @@ printThreadBlockage(StgTSO *tso)
     debugBelch("is blocked on an STM operation");
     break;
   default:
-    barf("printThreadBlockage: strange tso->why_blocked: %d for TSO %d (%p)",
-         tso->why_blocked, tso->id, tso);
+    barf("printThreadBlockage: strange tso->why_blocked: %d for TSO %ld (%p)",
+         tso->why_blocked, (long)tso->id, tso);
   }
 }
 
@@ -933,7 +950,7 @@ printAllThreads(void)
 
   debugBelch("all threads:\n");
 
-  for (i = 0; i < n_capabilities; i++) {
+  for (i = 0; i < getNumCapabilities(); i++) {
       cap = capabilities[i];
       debugBelch("threads on capability %d:\n", cap->no);
       for (t = cap->run_queue_hd; t != END_TSO_QUEUE; t = t->_link) {
