@@ -8,7 +8,7 @@
 
 #pragma once
 
-#include "rts/EventLogFormat.h"
+#include "eventlog/EventLog.h"
 #include "sm/NonMovingCensus.h"
 #include "Capability.h"
 
@@ -108,6 +108,8 @@ void traceEnd (void);
 void traceSchedEvent_ (Capability *cap, EventTypeNum tag,
                        StgTSO *tso, StgWord info1, StgWord info2);
 
+#define traceInitEvent(event) postInitEvent(event)
+
 /*
  * Record a GC event
  */
@@ -158,6 +160,11 @@ void traceEventGcStats_  (Capability *cap,
                           W_        par_tot_copied,
                           W_        par_balanced_copied);
 
+void traceEventMemReturn_  (Capability *cap,
+                          uint32_t    current_mblocks,
+                          uint32_t    needed_mblocks,
+                          uint32_t    returned_mblocks );
+
 /*
  * Record a spark event
  */
@@ -320,11 +327,19 @@ void traceConcUpdRemSetFlush(Capability *cap);
 void traceNonmovingHeapCensus(uint32_t log_blk_size,
                               const struct NonmovingAllocCensus *census);
 
+void traceIPE(StgInfoTable *info,
+               const char *table_name,
+               const char *closure_desc,
+               const char *ty_desc,
+               const char *label,
+               const char *module,
+               const char *srcloc );
 void flushTrace(void);
 
 #else /* !TRACING */
 
 #define traceSchedEvent(cap, tag, tso, other) /* nothing */
+#define traceInitEvent(event) /* nothing */
 #define traceSchedEvent2(cap, tag, tso, other, info) /* nothing */
 #define traceGcEvent(cap, tag) /* nothing */
 #define traceGcEventAtT(cap, ts, tag) /* nothing */
@@ -332,6 +347,7 @@ void flushTrace(void);
                            copied, slop, fragmentation, \
                            par_n_threads, par_max_copied, \
                            par_tot_copied, par_balanced_copied) /* nothing */
+#define traceEventMemReturn_(cap, current, needed, returned) /* nothing */
 #define traceHeapEvent(cap, tag, heap_capset, info1) /* nothing */
 #define traceEventHeapInfo_(heap_capset, gens, \
                             maxHeapSize, allocAreaSize, \
@@ -347,13 +363,14 @@ void flushTrace(void);
 #define traceCapEvent(cap, tag) /* nothing */
 #define traceCapsetEvent(tag, capset, info) /* nothing */
 #define traceWallClockTime_() /* nothing */
-#define traceOSProcessInfo_() /* nothing */
+#define traceOSProcessInfo_()  /* nothing */
 #define traceSparkCounters_(cap, counters, remaining) /* nothing */
 #define traceTaskCreate_(taskID, cap) /* nothing */
 #define traceTaskMigrate_(taskID, cap, new_cap) /* nothing */
 #define traceTaskDelete_(taskID) /* nothing */
 #define traceHeapProfBegin(profile_id) /* nothing */
 #define traceHeapProfCostCentre(ccID, label, module, srcloc, is_caf) /* nothing */
+#define traceIPE(info, table_name, closure_desc, ty_desc, label, module, srcloc) /* nothing */
 #define traceHeapProfSampleBegin(era) /* nothing */
 #define traceHeapBioProfSampleBegin(era, time) /* nothing */
 #define traceHeapProfSampleEnd(era) /* nothing */
@@ -374,7 +391,7 @@ void flushTrace(void);
 #endif /* TRACING */
 
 // If DTRACE is enabled, but neither DEBUG nor TRACING, we need a C land
-// wrapper for the user-msg probe (as we can't expand that in PrimOps.cmm)
+// wrapper for the user-msg probe (as we can't expand that in GHC.Builtin.PrimOpss.cmm)
 //
 #if !defined(DEBUG) && !defined(TRACING) && defined(DTRACE)
 
@@ -445,6 +462,8 @@ void dtraceUserMarkerWrapper(Capability *cap, char *msg);
                            par_max_copied,              \
                            par_balanced_copied,         \
                            par_tot_copied)
+#define dtraceEventMemReturn(current, needed, returned) \
+    HASKELLEVENT_MEM_RETURN(current, needed, returned)
 #define dtraceHeapInfo(heap_capset, gens,               \
                        maxHeapSize, allocAreaSize,      \
                        mblockSize, blockSize)           \
@@ -457,6 +476,8 @@ void dtraceUserMarkerWrapper(Capability *cap, char *msg);
                                 allocated)
 #define dtraceEventHeapSize(heap_capset, size)          \
     HASKELLEVENT_HEAP_SIZE(heap_capset, size)
+#define dtraceEventBlocksSize(heap_capset, size)        \
+    HASKELLEVENT_BLOCKS_SIZE(heap_capset, size)
 #define dtraceEventHeapLive(heap_capset, live)          \
     HASKELLEVENT_HEAP_LIVE(heap_capset, live)
 #define dtraceCapsetCreate(capset, capset_type)         \
@@ -516,12 +537,14 @@ void dtraceUserMarkerWrapper(Capability *cap, char *msg);
                            par_max_copied,              \
                            par_tot_copied,              \
                            par_balanced_copied)         /* nothing */
+#define dtraceEventMemReturn(current, needed, returned) /* nothing */
 #define dtraceHeapInfo(heap_capset, gens,               \
                        maxHeapSize, allocAreaSize,      \
                        mblockSize, blockSize)           /* nothing */
 #define dtraceEventHeapAllocated(cap, heap_capset,      \
                                  allocated)             /* nothing */
 #define dtraceEventHeapSize(heap_capset, size)          /* nothing */
+#define dtraceEventBlocksSize(heap_capset, size)        /* nothing */
 #define dtraceEventHeapLive(heap_capset, live)          /* nothing */
 #define dtraceCapCreate(cap)                            /* nothing */
 #define dtraceCapDelete(cap)                            /* nothing */
@@ -617,6 +640,18 @@ INLINE_HEADER void traceCapDisable(Capability *cap STG_UNUSED)
 {
     traceCapEvent(cap, EVENT_CAP_DISABLE);
     dtraceCapDisable((EventCapNo)cap->no);
+
+    // Ensure that the eventlog buffer is flushed since otherwise its events
+    // may never make it to the output stream.
+    // See Note [Eventlog concurrency].
+#if defined(TRACING)
+    if (eventlog_enabled) {
+        flushLocalEventsBuf(cap);
+    }
+# else
+    flushLocalEventsBuf(cap);
+#endif
+
 }
 
 INLINE_HEADER void traceEventThreadWakeup(Capability *cap       STG_UNUSED,
@@ -723,6 +758,17 @@ INLINE_HEADER void traceEventGcStats(Capability *cap            STG_UNUSED,
                        par_tot_copied, par_balanced_copied);
 }
 
+INLINE_HEADER void traceEventMemReturn(Capability *cap            STG_UNUSED,
+                                     uint32_t    current_mblocks STG_UNUSED,
+                                     uint32_t    needed_mblocks  STG_UNUSED,
+                                     uint32_t    returned_mblocks STG_UNUSED)
+{
+    if (RTS_UNLIKELY(TRACE_gc)) {
+        traceEventMemReturn_(cap, current_mblocks, needed_mblocks, returned_mblocks);
+    }
+    dtraceEventMemReturn(current_mblocks, needed_mblocks, returned_mblocks);
+}
+
 INLINE_HEADER void traceEventHeapInfo(CapsetID    heap_capset   STG_UNUSED,
                                       uint32_t  gens          STG_UNUSED,
                                       W_        maxHeapSize   STG_UNUSED,
@@ -756,6 +802,14 @@ INLINE_HEADER void traceEventHeapSize(Capability *cap         STG_UNUSED,
     dtraceEventHeapSize(heap_capset, heap_size);
 }
 
+INLINE_HEADER void traceEventBlocksSize(Capability *cap         STG_UNUSED,
+                                        CapsetID    heap_capset STG_UNUSED,
+                                        W_        heap_size   STG_UNUSED)
+{
+    traceHeapEvent(cap, EVENT_BLOCKS_SIZE, heap_capset, heap_size);
+    dtraceEventBlocksSize(heap_capset, heap_size);
+}
+
 INLINE_HEADER void traceEventHeapLive(Capability *cap         STG_UNUSED,
                                       CapsetID    heap_capset STG_UNUSED,
                                       W_        heap_live   STG_UNUSED)
