@@ -1095,7 +1095,7 @@ resetMutableObjects(traverseState* ts)
         // Traversing through mut_list is necessary
         // because we can find MUT_VAR objects which have not been
         // visited during heap traversal.
-        for (n = 0; n < n_capabilities; n++) {
+        for (n = 0; n < getNumCapabilities(); n++) {
           for (bd = capabilities[n]->mut_lists[g]; bd != NULL; bd = bd->link) {
             for (ml = bd->start; ml < bd->free; ml++) {
                 traverseMaybeInitClosureData(ts, (StgClosure *)*ml);
@@ -1239,7 +1239,6 @@ inner_loop:
         traversePushClosure(ts, (StgClosure *) tso->trec, c, sep, child_data);
         if (   tso->why_blocked == BlockedOnMVar
                || tso->why_blocked == BlockedOnMVarRead
-               || tso->why_blocked == BlockedOnIOCompletion
                || tso->why_blocked == BlockedOnBlackHole
                || tso->why_blocked == BlockedOnMsgThrowTo
             ) {
