@@ -159,24 +159,40 @@ StgWord8 the_gc_thread[sizeof(gc_thread) + 64 * sizeof(gen_workspace)]
 
 /* Note [n_gc_threads]
 This is a global variable that originally tracked the number of threads
-participating in the current gc. It's meaing has diverged from this somewhate.
-In practise, it now takes one of the values {1, n_capabilities}. Don't be
-tricked into thinking this means garbage collections must have 1 or
-n_capabilities participating: idle capabilities (idle_cap[cap->no]) are included
-in the n_gc_thread count.
+participating in the current gc. It's meaning has diverged from this somewhat,
+as it does not distinguish betweeen idle and non-idle threads. An idle thread
+is a thread i where `idle[i]` is true, using the idle array passed to
+GarbageCollect.
+
+In practice, it now takes one of the values {1, n_capabilities}, when a SYNC_GC_SEQ
+is requested it takes 1, when a SYNC_GC_PAR is requested it takes n_capabilities.
 
 Clearly this is in need of some tidying up, but for now we tread carefully. We
-check n_gc_threads > 1 to see whether we are in a parallel or sequential. We
-ensure n_gc_threads > 1 before iterating over gc_threads a la:
-
-for(i=0;i<n_gc_threads;++i) { foo(gc_thread[i]; )}
+call is_par_gc() to see whether we are in a parallel or sequential collection.
+If we are in a parallel collection we iterate over gc_threads, being careful to
+account for idle caps. If we are in a sequential collection we deal only with
+the thread local gct.
+Of course this is valid only inside GarbageCollect ().
 
 Omitting this check has led to issues such as #19147.
 */
+
+static bool is_par_gc(void);
+
 uint32_t n_gc_threads;
 static uint32_t n_gc_idle_threads;
 bool work_stealing;
 
+static bool is_par_gc() {
+#if defined(THREADED_RTS)
+    if(n_gc_threads == 1) { return false; }
+    ASSERT(n_gc_threads > n_gc_idle_threads);
+    return n_gc_threads - n_gc_idle_threads > 1;
+#else
+    return false;
+#endif
+}
+
 // For stats:
 static long copied;        // *words* copied & scavenged during this GC
 
@@ -275,6 +291,7 @@ GarbageCollect (uint32_t collect_gen,
       any_work, scav_find_work, max_n_todo_overflow;
 #if defined(THREADED_RTS)
   gc_thread *saved_gct;
+  bool gc_sparks_all_caps;
 #endif
   uint32_t g, n;
   // The time we should report our heap census as occurring at, if necessary.
@@ -292,7 +309,7 @@ GarbageCollect (uint32_t collect_gen,
 #endif
 
 #if defined(PROFILING)
-  CostCentreStack *save_CCS[n_capabilities];
+  CostCentreStack *save_CCS[getNumCapabilities()];
 #endif
 
   ACQUIRE_SM_LOCK;
@@ -323,7 +340,7 @@ GarbageCollect (uint32_t collect_gen,
 
   // attribute any costs to CCS_GC
 #if defined(PROFILING)
-  for (n = 0; n < n_capabilities; n++) {
+  for (n = 0; n < getNumCapabilities(); n++) {
       save_CCS[n] = capabilities[n]->r.rCCCS;
       capabilities[n]->r.rCCCS = CCS_GC;
   }
@@ -358,7 +375,8 @@ GarbageCollect (uint32_t collect_gen,
           static_flag == STATIC_FLAG_A ? STATIC_FLAG_B : STATIC_FLAG_A;
   }
 
-  if (major_gc) {
+  /* N.B. We currently don't unload code with the non-moving collector. */
+  if (major_gc && !RtsFlags.GcFlags.useNonmoving) {
       unload_mark_needed = prepareUnloadCheck();
   } else {
       unload_mark_needed = false;
@@ -379,9 +397,9 @@ GarbageCollect (uint32_t collect_gen,
    * here
   */
   if (gc_type == SYNC_GC_PAR) {
-      n_gc_threads = n_capabilities;
+      n_gc_threads = getNumCapabilities();
       n_gc_idle_threads = 0;
-      for (uint32_t i = 0; i < n_capabilities; ++i) {
+      for (uint32_t i = 0; i < getNumCapabilities(); ++i) {
           if (idle_cap[i]) {
               ASSERT(i != gct->thread_index);
               ++n_gc_idle_threads;
@@ -389,14 +407,11 @@ GarbageCollect (uint32_t collect_gen,
       }
   } else {
       n_gc_threads = 1;
-      n_gc_idle_threads = n_capabilities - 1;
+      n_gc_idle_threads = getNumCapabilities() - 1;
   }
   work_stealing = RtsFlags.ParFlags.parGcLoadBalancingEnabled &&
       N >= RtsFlags.ParFlags.parGcLoadBalancingGen &&
-      (gc_type == SYNC_GC_PAR) &&
-      (n_gc_threads - n_gc_idle_threads) > 1;
-      // ^ if we are the only non-idle gc thread, turn off work stealing
-      //
+      is_par_gc();
       // It's not always a good idea to do load balancing in parallel
       // GC.  In particular, for a parallel program we don't want to
       // lose locality by moving cached data into another CPU's cache
@@ -408,22 +423,22 @@ GarbageCollect (uint32_t collect_gen,
       // a flag.
 #else
   n_gc_threads = 1;
-  n_gc_idle_threads = 0;
   work_stealing = false;
+  n_gc_idle_threads = 0;
 #endif
 
   SEQ_CST_STORE(&gc_running_threads, 0);
 
   ASSERT(n_gc_threads > 0);
-  ASSERT(n_gc_threads <= n_capabilities);
-  ASSERT(n_gc_idle_threads < n_capabilities);
+  ASSERT(n_gc_threads <= getNumCapabilities());
+  ASSERT(n_gc_idle_threads < getNumCapabilities());
   // If we are work stealing, there better be another(i.e. not us) non-idle gc
   // thread
   ASSERT(!work_stealing || n_gc_threads - 1 > n_gc_idle_threads);
 
 
   debugTrace(DEBUG_gc, "GC (gen %d, using %d thread(s), %s work stealing)",
-             N, (int)n_capabilities - (int)n_gc_idle_threads,
+             N, (int)getNumCapabilities() - (int)n_gc_idle_threads,
              work_stealing ? "with": "without");
 
 #if defined(DEBUG)
@@ -484,8 +499,8 @@ GarbageCollect (uint32_t collect_gen,
   // of markSomeCapabilities() because markSomeCapabilities() can only
   // call back into the GC via mark_root() (due to the gct register
   // variable).
-  if (n_gc_threads == 1) {
-      for (n = 0; n < n_capabilities; n++) {
+  if (!is_par_gc()) {
+      for (n = 0; n < getNumCapabilities(); n++) {
 #if defined(THREADED_RTS)
           scavenge_capability_mut_Lists1(capabilities[n]);
 #else
@@ -494,7 +509,7 @@ GarbageCollect (uint32_t collect_gen,
       }
   } else {
       scavenge_capability_mut_lists(gct->cap);
-      for (n = 0; n < n_capabilities; n++) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           if (idle_cap[n]) {
               markCapability(mark_root, gct, capabilities[n],
                              true/*don't mark sparks*/);
@@ -509,8 +524,8 @@ GarbageCollect (uint32_t collect_gen,
 
   // follow all the roots that the application knows about.
   gct->evac_gen_no = 0;
-  if (n_gc_threads == 1) {
-      for (n = 0; n < n_capabilities; n++) {
+  if (!is_par_gc()) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           markCapability(mark_root, gct, capabilities[n],
                          true/*don't mark sparks*/);
       }
@@ -542,6 +557,9 @@ GarbageCollect (uint32_t collect_gen,
   StgTSO *resurrected_threads = END_TSO_QUEUE;
   // must be last...  invariant is that everything is fully
   // scavenged at this point.
+#if defined(THREADED_RTS)
+  gc_sparks_all_caps = !work_stealing || !is_par_gc();
+#endif
   work_stealing = false;
   while (traverseWeakPtrList(&dead_weak_ptr_list, &resurrected_threads))
   {
@@ -554,12 +572,13 @@ GarbageCollect (uint32_t collect_gen,
   gcStableNameTable();
 
 #if defined(THREADED_RTS)
-  if (n_gc_threads == 1) {
-      for (n = 0; n < n_capabilities; n++) {
+  // See Note [Pruning the spark pool]
+  if (gc_sparks_all_caps) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           pruneSparkQueue(false, capabilities[n]);
       }
   } else {
-      for (n = 0; n < n_capabilities; n++) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           if (n == cap->no || idle_cap[n]) {
               pruneSparkQueue(false, capabilities[n]);
          }
@@ -602,11 +621,16 @@ GarbageCollect (uint32_t collect_gen,
       const gc_thread* thread;
 
       // see Note [n_gc_threads]
-      if (n_gc_threads > 1) {
+      if (is_par_gc()) {
+          int other_active_threads = n_gc_threads - n_gc_idle_threads - 1;
+          ASSERT(other_active_threads > 0);
+
           for (i=0; i < n_gc_threads; i++) {
+              if(idle_cap[i]) { continue; }
               copied += RELAXED_LOAD(&gc_threads[i]->copied);
           }
           for (i=0; i < n_gc_threads; i++) {
+              if(idle_cap[i]) { continue; }
               thread = gc_threads[i];
               debugTrace(DEBUG_gc,"thread %d:", i);
               debugTrace(DEBUG_gc,"   copied           %ld",
@@ -624,14 +648,18 @@ GarbageCollect (uint32_t collect_gen,
 
               par_max_copied = stg_max(RELAXED_LOAD(&thread->copied), par_max_copied);
               par_balanced_copied_acc +=
-                  stg_min(n_gc_threads * RELAXED_LOAD(&thread->copied), copied);
+                  stg_min((other_active_threads + 1) * RELAXED_LOAD(&thread->copied), copied);
           }
+
           // See Note [Work Balance] for an explanation of this computation
           par_balanced_copied =
-              (par_balanced_copied_acc - copied + (n_gc_threads - 1) / 2) /
-              (n_gc_threads - 1);
+              (par_balanced_copied_acc - copied + other_active_threads / 2) /
+              other_active_threads;
       } else {
           copied += gct->copied;
+          any_work += gct->any_work;
+          scav_find_work += gct->scav_find_work;
+          max_n_todo_overflow += gct->max_n_todo_overflow;
       }
   }
 
@@ -649,14 +677,14 @@ GarbageCollect (uint32_t collect_gen,
 
     if (g == N) {
       generations[g].collections++; // for stats
-      if (n_gc_threads > 1) generations[g].par_collections++;
+      if (is_par_gc()) generations[g].par_collections++;
     }
 
     // Count the mutable list as bytes "copied" for the purposes of
     // stats.  Every mutable list is copied during every GC.
     if (g > 0) {
         W_ mut_list_size = 0;
-        for (n = 0; n < n_capabilities; n++) {
+        for (n = 0; n < getNumCapabilities(); n++) {
             mut_list_size += countOccupied(capabilities[n]->mut_lists[g]);
         }
         copied +=  mut_list_size;
@@ -809,7 +837,7 @@ GarbageCollect (uint32_t collect_gen,
     // add in the partial blocks in the gen_workspaces
     {
         uint32_t i;
-        for (i = 0; i < n_capabilities; i++) {
+        for (i = 0; i < getNumCapabilities(); i++) {
             live_words  += gcThreadLiveWords(i, gen->no);
             live_blocks += gcThreadLiveBlocks(i, gen->no);
         }
@@ -819,11 +847,9 @@ GarbageCollect (uint32_t collect_gen,
   // Flush the update remembered sets. See Note [Eager update remembered set
   // flushing] in NonMovingMark.c
   if (RtsFlags.GcFlags.useNonmoving) {
-      RELEASE_SM_LOCK;
-      for (n = 0; n < n_capabilities; n++) {
-          nonmovingAddUpdRemSetBlocks(&capabilities[n]->upd_rem_set.queue);
+      for (n = 0; n < getNumCapabilities(); n++) {
+          nonmovingAddUpdRemSetBlocks(&capabilities[n]->upd_rem_set);
       }
-      ACQUIRE_SM_LOCK;
   }
 
   // Mark and sweep the oldest generation.
@@ -844,8 +870,6 @@ GarbageCollect (uint32_t collect_gen,
       // old_weak_ptr_list should be empty.
       ASSERT(oldest_gen->old_weak_ptr_list == NULL);
 
-      // we may need to take the lock to allocate mark queue blocks
-      RELEASE_SM_LOCK;
       // dead_weak_ptr_list contains weak pointers with dead keys. Those need to
       // be kept alive because we'll use them in finalizeSchedulers(). Similarly
       // resurrected_threads are also going to be used in resurrectedThreads()
@@ -855,10 +879,9 @@ GarbageCollect (uint32_t collect_gen,
 #if !defined(THREADED_RTS)
       // In the non-threaded runtime this is the only time we push to the
       // upd_rem_set
-      nonmovingAddUpdRemSetBlocks(&gct->cap->upd_rem_set.queue);
+      nonmovingAddUpdRemSetBlocks(&gct->cap->upd_rem_set);
 #endif
       nonmovingCollect(&dead_weak_ptr_list, &resurrected_threads);
-      ACQUIRE_SM_LOCK;
   }
 
   // Update the max size of older generations after a major GC:
@@ -1040,7 +1063,7 @@ GarbageCollect (uint32_t collect_gen,
 
   // restore enclosing cost centre
 #if defined(PROFILING)
-  for (n = 0; n < n_capabilities; n++) {
+  for (n = 0; n < getNumCapabilities(); n++) {
       capabilities[n]->r.rCCCS = save_CCS[n];
   }
 #endif
@@ -1189,7 +1212,7 @@ freeGcThreads (void)
     if (gc_threads != NULL) {
 #if defined(THREADED_RTS)
         uint32_t i;
-        for (i = 0; i < n_capabilities; i++) {
+        for (i = 0; i < getNumCapabilities(); i++) {
             for (g = 0; g < RtsFlags.GcFlags.generations; g++)
             {
                 freeWSDeque(gc_threads[i]->gens[g].todo_q);
@@ -1227,7 +1250,6 @@ inc_running (void)
     // See Note [Synchronising work stealing]
     StgWord new;
     new = atomic_inc(&gc_running_threads, 1);
-    ASSERT(new <= n_gc_threads);
     return new;
 }
 
@@ -1289,7 +1311,7 @@ scavenge_until_all_done (void)
 
     for(;;) {
 #if defined(THREADED_RTS)
-        if (n_gc_threads > 1) {
+        if (is_par_gc()) {
             scavenge_loop();
         } else {
             scavenge_loop1();
@@ -1313,7 +1335,7 @@ scavenge_until_all_done (void)
         // If there's no hope of stealing more work, then there's nowhere else
         // work can come from and we are finished
 #if defined(THREADED_RTS)
-        if(n_gc_threads > 1 && work_stealing && r != 0) {
+        if(is_par_gc() && work_stealing && r != 0) {
             NONATOMIC_ADD(&gct->any_work, 1);
             ACQUIRE_LOCK(&gc_running_mutex);
             // this is SEQ_CST because I haven't considered if it could be
@@ -1350,7 +1372,6 @@ void
 gcWorkerThread (Capability *cap)
 {
     gc_thread *saved_gct;
-
     // necessary if we stole a callee-saves register for gct:
     saved_gct = gct;
 
@@ -1381,13 +1402,10 @@ gcWorkerThread (Capability *cap)
     scavenge_until_all_done();
 
 #if defined(THREADED_RTS)
-    // Now that the whole heap is marked, we discard any sparks that
-    // were found to be unreachable.  The main GC thread is currently
-    // marking heap reachable via weak pointers, so it is
-    // non-deterministic whether a spark will be retained if it is
-    // only reachable via weak pointers.  To fix this problem would
-    // require another GC barrier, which is too high a price.
-    pruneSparkQueue(false, cap);
+    // See Note [Pruning the spark pool]
+    if(work_stealing && is_par_gc()) {
+        pruneSparkQueue(false, cap);
+    }
 #endif
 
     // Wait until we're told to continue
@@ -1420,26 +1438,26 @@ void
 waitForGcThreads (Capability *cap, bool idle_cap[])
 {
     // n_gc_threads is not valid here, we're too early
-    uint32_t n_threads = n_capabilities;
+    uint32_t n_threads = getNumCapabilities();
     const uint32_t me = cap->no;
     uint32_t i, cur_n_gc_entered;
     Time t0, t1, t2;
 
     t0 = t1 = t2 = getProcessElapsedTime();
 
-    for(i = 0; i < n_capabilities; ++i) {
+    for(i = 0; i < getNumCapabilities(); ++i) {
         if (i == me || idle_cap[i]) {
             --n_threads;
         }
     }
 
-    ASSERT(n_threads < n_capabilities); // must be less because we don't count ourself
+    ASSERT(n_threads < getNumCapabilities()); // must be less because we don't count ourself
     if(n_threads == 0) { return; }
 
     ACQUIRE_LOCK(&gc_entry_mutex);
     while((cur_n_gc_entered = SEQ_CST_LOAD(&n_gc_entered)) != n_threads) {
         ASSERT(cur_n_gc_entered < n_threads);
-        for(i = 0; i < n_capabilities; ++i) {
+        for(i = 0; i < getNumCapabilities(); ++i) {
             if (i == me || idle_cap[i]) { continue; }
             if (SEQ_CST_LOAD(&gc_threads[i]->wakeup) != GC_THREAD_STANDING_BY) {
                 prodCapability(capabilities[i], cap->running_task);
@@ -1480,7 +1498,7 @@ wakeup_gc_threads (uint32_t me USED_IF_THREADS,
 #if defined(THREADED_RTS)
     uint32_t i;
 
-    if (n_gc_threads == 1) return;
+    if (!is_par_gc()) return;
 
 #if defined(DEBUG)
     StgWord num_idle = 0;
@@ -1516,7 +1534,7 @@ shutdown_gc_threads (uint32_t me USED_IF_THREADS USED_IF_DEBUG,
 {
 #if defined(THREADED_RTS)
 
-    if (n_gc_threads == 1) return;
+    if (!is_par_gc()) return;
 
     // we need to wait for `n_threads` threads. -1 because that's ourself
     StgInt n_threads = (StgInt)n_gc_threads - 1 - (StgInt)n_gc_idle_threads;
@@ -1529,7 +1547,7 @@ shutdown_gc_threads (uint32_t me USED_IF_THREADS USED_IF_DEBUG,
     }
 #if defined(DEBUG)
     uint32_t i;
-    for (i=0; i < n_capabilities; i++) {
+    for (i=0; i < getNumCapabilities(); i++) {
         if (i == me || idle_cap[i]) continue;
         ASSERT(SEQ_CST_LOAD(&gc_threads[i]->wakeup) == GC_THREAD_WAITING_TO_CONTINUE);
     }
@@ -1542,7 +1560,7 @@ shutdown_gc_threads (uint32_t me USED_IF_THREADS USED_IF_DEBUG,
 void
 releaseGCThreads (Capability *cap USED_IF_THREADS, bool idle_cap[])
 {
-    const uint32_t n_threads = n_capabilities;
+    const uint32_t n_threads = getNumCapabilities();
     const uint32_t me = cap->no;
     uint32_t i;
 #if defined(DEBUG)
@@ -1594,14 +1612,14 @@ prepare_collected_gen (generation *gen)
 
     if (RtsFlags.GcFlags.useNonmoving && g == oldest_gen->no) {
         // Nonmoving heap's mutable list is always a root.
-        for (i = 0; i < n_capabilities; i++) {
+        for (i = 0; i < getNumCapabilities(); i++) {
             stash_mut_list(capabilities[i], g);
         }
     } else if (g != 0) {
         // Otherwise throw away the current mutable list. Invariant: the
         // mutable list always has at least one block; this means we can avoid
         // a check for NULL in recordMutable().
-        for (i = 0; i < n_capabilities; i++) {
+        for (i = 0; i < getNumCapabilities(); i++) {
             bdescr *old = RELAXED_LOAD(&capabilities[i]->mut_lists[g]);
             freeChain(old);
 
@@ -1638,7 +1656,7 @@ prepare_collected_gen (generation *gen)
 
     // grab all the partial blocks stashed in the gc_thread workspaces and
     // move them to the old_blocks list of this gen.
-    for (n = 0; n < n_capabilities; n++) {
+    for (n = 0; n < getNumCapabilities(); n++) {
         ws = &gc_threads[n]->gens[gen->no];
 
         for (bd = ws->part_list; bd != NULL; bd = next) {
@@ -1737,7 +1755,7 @@ prepare_uncollected_gen (generation *gen)
     // save the current mutable lists for this generation, and
     // allocate a fresh block for each one.  We'll traverse these
     // mutable lists as roots early on in the GC.
-    for (i = 0; i < n_capabilities; i++) {
+    for (i = 0; i < getNumCapabilities(); i++) {
         stash_mut_list(capabilities[i], gen->no);
     }
 
@@ -1811,7 +1829,7 @@ collect_pinned_object_blocks (void)
     const bool use_nonmoving = RtsFlags.GcFlags.useNonmoving;
     generation *const gen = (use_nonmoving && major_gc) ? oldest_gen : g0;
 
-    for (uint32_t n = 0; n < n_capabilities; n++) {
+    for (uint32_t n = 0; n < getNumCapabilities(); n++) {
         bdescr *last = NULL;
         if (use_nonmoving && gen == oldest_gen) {
             // Mark objects as belonging to the nonmoving heap
@@ -1927,7 +1945,7 @@ resizeGenerations (void)
     // minimum size for generation zero
     min_alloc = stg_max((RtsFlags.GcFlags.pcFreeHeap * max) / 200,
                         RtsFlags.GcFlags.minAllocAreaSize
-                        * (W_)n_capabilities);
+                        * (W_)getNumCapabilities());
 
     // Auto-enable compaction when the residency reaches a
     // certain percentage of the maximum heap size (default: 30%).
@@ -1999,7 +2017,7 @@ static void
 resize_nursery (void)
 {
     const StgWord min_nursery =
-      RtsFlags.GcFlags.minAllocAreaSize * (StgWord)n_capabilities;
+      RtsFlags.GcFlags.minAllocAreaSize * (StgWord)getNumCapabilities();
 
     if (RtsFlags.GcFlags.generations == 1)
     {   // Two-space collector:
