@@ -26,7 +26,8 @@ enum SweepResult {
 };
 
 // Determine which list a marked segment should be placed on and initialize
-// next_free indices as appropriate.
+// next_free indices as appropriate. Additionally, we must clear the mark
+// bitmap entries associated with swept blocks.
 GNUC_ATTR_HOT static enum SweepResult
 nonmovingSweepSegment(struct NonmovingSegment *seg)
 {
@@ -39,6 +40,8 @@ nonmovingSweepSegment(struct NonmovingSegment *seg)
         if (seg->bitmap[i] == nonmovingMarkEpoch) {
             found_live = true;
         } else if (!found_free) {
+            // This is the first free block we've found; set next_free,
+            // next_free_snap, and the scan pointer.
             found_free = true;
             seg->next_free = i;
             nonmovingSegmentInfo(seg)->next_free_snap = i;
@@ -64,7 +67,6 @@ nonmovingSweepSegment(struct NonmovingSegment *seg)
     } else {
         ASSERT(seg->next_free == 0);
         ASSERT(nonmovingSegmentInfo(seg)->next_free_snap == 0);
-        nonmovingClearBitmap(seg);
         return SEGMENT_FREE;
     }
 }
@@ -277,7 +279,7 @@ dirty_BLOCKING_QUEUE:
 /* N.B. This happens during the pause so we own all capabilities. */
 void nonmovingSweepMutLists()
 {
-    for (uint32_t n = 0; n < n_capabilities; n++) {
+    for (uint32_t n = 0; n < getNumCapabilities(); n++) {
         Capability *cap = capabilities[n];
         bdescr *old_mut_list = cap->mut_lists[oldest_gen->no];
         cap->mut_lists[oldest_gen->no] = allocBlockOnNode_lock(cap->node);
